TITLE Forth-98 system

PAGE 62,132	;62 lines per page, 132 characters per line

;===============================================================
;
;	8098 eForth 1.0 by Dingkezhong,1991
;
;	This eForth system was developed using chipForth from Forth, Inc.
;	and tested on a ATD/CJ-98 microcomputer developing system.
;	The eForth Model was developed by Bill Muench and C. H. Ting.
;
;	The goal of this implementation is to port the eForth Model
;	to a ROM based 16 bit microprocessor, Intel 8098.
;	Deviations from the original eForth Model are:
;
;		All kernel words are assembled as DB statements.
;		Memory map is tailored to a ROM based system.
;		$COLON and $USER are modified to compile LCALL doLIST.
;		call, compiles a LCALL with a relative offset address.
;		USER, VARIABLE and : are modified to use above 'call,'.
;		FORTH vocabulary pointer is a pair user variables.
;		BYE is deleted.
;
;	To assemble this source file and generate a ROM image,
;	type the following commands using MASM and LINK:
;		>MASM 8098;
;		>LINK 8098;
;	The resulting 8098.EXE contains the binary image suitable
;	for EPROM programming.	The actual image is offset by 200H
;	bytes from the beginning of the .EXE file.  This image
;	must be placed in a PROM from 0 to 1FFFH, and it uses a RAM
;	chip from 8000H to 9FFFH.  If your system does not have
;	this memory configuration, modify the memory pointers in
;	the source file accordingly.  Places to be modified are
;	marked by '******'.
;
;	Considering that most words are in high level.Although 8098 is a
;	fast processor,you cannot get great performance of this implementation.
;	You are encouraged to recode some of the high level words
;	to optimize its performance.
;
;	Direct your questions and contributions to:
;
;		Dingkezhong
;		PO.box 0889
;		Southeast University
;		Nanjing 210018,PRC.
;
;===============================================================


;; Version control

VER		EQU	1			;major release version
EXT		EQU	0			;minor extension

;; Constants

TRUEE		EQU	-1			;true flag

COMPO		EQU	040H			;lexicon compile only bit
IMEDD		EQU	080H			;lexicon immediate bit
MASKK		EQU	07F1FH			;lexicon bit mask

CELLL		EQU	2			;size of a cell
BASEE		EQU	10			;default radix
VOCSS		EQU	8			;depth of vocabulary stack

BKSPP		EQU	8			;back space
LF		EQU	10			;line feed
CRR		EQU	13			;carriage return
FSRM		EQU	19			;file sending request message
CTRL_Z          EQU     26                      ;send process over message
ERR		EQU	27			;error escape
TIC		EQU	39			;tick

LCALLL		EQU	0EFFDH			;NOP LCALL opcodes

;; Memory allocation

RESET		EQU	0080H			;emulate 8098 reset cell  *******
EM		EQU	0E000H			;top of memory	******
BM		EQU	0000H			;bottom of ROM	******
TOPRAM		EQU	0C000H			;top of RAM (all) ******
BMRAM		EQU	2000H			;bottom of RAM (all) ******
UBMRAM		EQU	0A000H			;bottom of RAM ******
US              EQU     128*CELLL               ;user area size in cells
RTS		EQU	128*CELLL		;return stack/TIB size
RESERV		EQU	100H			;8098 reserved in cells *****
NNPP		EQU	UBMRAM-CELLL		;next available Name Pointer
CCPP		EQU	BMRAM+RESERV		;next available Code Pointer

;/RAM area

UPP		EQU	TOPRAM-US		;start of user area (UP0)
RPP		EQU	UPP-8*CELLL		;start of return stack (RP0)
TIBB		EQU	RPP-RTS 		;terminal input buffer (TIB)
SPP		EQU	TIBB-8*CELLL		;start of data stack (SP0)
VPP0		EQU	UBMRAM			;bottom of RAM ******
CRVPP0		EQU	VPP0+0CH		;current available RAM ***

;/ROM area

COLDD		EQU	0100H			;cold start vector *******
USEE            EQU     COLDD+80H               ;system user variable*****
CODEE           EQU     USEE+US                 ;code dictionary  *****
NAMEE		EQU	EM-CELLL		;name dictionary   *******

;; Serial I/O devices

CTLPORT 	EQU	0E001H			 ;CC8251 control port  ***
DATPORT 	EQU	0E000H			 ;CC8251 data prort    ***

;; Parallel I/O ports

APORT           EQU     0FFFCH                   ;8255  PORT A.
BPORT           EQU     0FFFDH                   ;8255  PORT B.
CPORT           EQU     0FFFEH                   ;8255  PORT C.
CNTL            EQU     0FFFFH                   ;8255  CONTROL


;; 8098 register usage:

;AX		 EQU	 20H
;BX		 EQU	 22H
;CX		 EQU	 24H
;DX		 EQU	 26H
;VP		 EQU	 28H
;SP		 EQU	 18H
;RP		 EQU	 1CH
;IP		 EQU	 1AH
;UP		 EQU	 1EH

;; Initialize assembly variables

_LINK	= 0					;force a null link
_NAME	= NAMEE 				;initialize name pointer
_CODE	= CODEE 				;initialize code pointer
_USER	= 4*CELLL				;first user variable offset

;; Define assembly macros

;	Compile a code definition header.

$CODE	MACRO	LEX,NAME,LABEL
	EVEN					;;force to cell boundary
LABEL:						;;assembly label
	_CODE	= $				;;save code pointer
	_LEN	= (LEX AND 01FH)/CELLL		;;string cell count, round down
	_NAME	= _NAME-((_LEN+3)*CELLL)	;;new header on cell boundary
ORG	_NAME					;;set name pointer
	DW	 _CODE,_LINK			;;token pointer and link
	_LINK	= $				;;link points to a name string
	DB	LEX,NAME			;;name string
ORG	_CODE					;;restore code pointer
	ENDM

;	Compile a colon definition header.

$COLON	MACRO	LEX,NAME,LABEL
	$CODE	LEX,NAME,LABEL
	DB	0FDH				 ;NOP
	DB	0EFH				 ;LCALL DOLST
	DW	DOLST-_CODE-4
	ENDM

;	Compile a user variable header.

$USER	MACRO	LEX,NAME,LABEL
	$CODE	LEX,NAME,LABEL
	DB	0FDH				 ;NOP
	DB	0EFH				 ;LCALL DOUSE
	DW	DOUSE-_CODE-4
	DW	_USER
	_USER = _USER+CELLL
	ENDM
$CONST  MACRO   LEX,NAME,LABEL
        $CODE   LEX,NAME,LABEL
        DB      0FDH                            ; NOP
        DB      0EFH                            ; LCALL DOCONST
        DW      DOCONST-_CODE-4
        ENDM

$VAR	MACRO	LEX,NAME,LABEL
	$CODE	LEX,NAME,LABEL
	DB	0FDH				;NOP
	DB	0EFH				;LCALL DOVAR
	DW	DOVAR-_CODE-4
	ENDM

$DOES	MACRO	LEX,NAME,LABEL
	$CODE	LEX,NAME,LABEL
	DB	0FDH				;NOP
	DB	0EFH				;LCALL DODOES
	DW	DODOES-_CODE-4
	ENDM

;	Assemble inline direct threaded code ending.

$NEXT	MACRO
	DB     0A2H,1BH,20H			;LD AX,[IP]+
	DB     0E3H,20H 			;BR [AX]
	ENDM

;; Main entry points and COLD start data

MAIN	SEGMENT
	ASSUME	CS:MAIN,DS:MAIN,ES:MAIN,SS:MAIN
ORG	0000H
ORIG:
		DW	9 DUP ( RESET ) 	;interrupt vector
ORG	0012H
		DB	6 DUP ( 0FFH )		;reserved
ORG	0018H
		DB	10111101B		;CCR chip control
ORG	0019H
		DB	0FFH			;reserved
ORG	001CH
		DB	100 DUP ( 0FFH )	;reserved

ORG	RESET					;reset area

		DB	0E7H			;LJMP COLDD
		DW	COLDD-2083H

ORG	COLDD					;beginning of cold boot area
		DB	0B1H,05H,23H		;LDB BH,#05H
DLAY:           DB      0B1H,00H,22H            ;LDB BL,#00H
DLAY1:          DB      0FDH                    ;NOP
                DB      0E0H,22H,0FCH           ;DJNZ BL,DLAY1
                DB      0E0H,23H,0F6H           ;DJNZ BH,DLAY
                DB      0A1H                    ;LD SP,SPP
		DW	SPP
		DB	18H
		DB	0A1H			;LD RP,RPP
		DW	RPP
		DB	1CH
		DB	0A1H			;LD UP,UPP
		DW	UPP
		DB	1EH
		DB	0A1H			;LD VP,VPP0
		DW	CRVPP0
		DB	28H
		DB	0A1H			;LD IP,COLD1
		DW	COLD1
		DB	1AH
                DB      0B1H,80H,20H            ;LDB AL,#80H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,CNTL[0]
                DB      0B1H,8FH,20H            ;LDB AL,#8FH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,CPORT[0]
		DB	0B1H,4EH,20H		;LDB AL,#4EH
		DB	0C7H,01H,01H,0E0H,20H	;STB AL,CTLPORT[0]
		DB	0B1H,15H,20H		;LDB AL,#15H
		DB	0C7H,01H,01H,0E0H,20H	;STB AL,CTLPORT[0]
		$NEXT				;to high level cold start

; COLD start moves the following to USER variables.
; MUST BE IN SAME ORDER AS USER VARIABLES.

ORG     USEE

EVEN						;align to cell boundary

UZERO:		DW	4 DUP (0)		;reserved space in user area
		DW	SPP			;SP0
		DW	RPP			;RP0
                DW      UPP                     ;UP0
		DW	QRX			;'?KEY
		DW	TXSTO			;'EMIT
		DW	ACCEP			;'EXPECT
		DW	KTAP			;'TAP
		DW	TXSTO			;'ECHO
		DW	DOTOK			;'PROMPT
		DW	BASEE			;BASE
		DW	0			;tmp
		DW	0			;SPAN
		DW	0			;>IN
		DW	0			;#TIB
		DW	TIBB			;TIB
		DW	0			;CSP
		DW	INTER			;'EVAL
		DW	NUMBQ			;'NUMBER
		DW	0			;HLD
		DW	0			;HANDLER
		DW	0			;VOC-LINK
		DW	0			;CONTEXT pointer
		DW	VOCSS+1 DUP (0) 	;vocabulary stack
		DW	0			;CURRENT pointer
		DW	0			;vocabulary link pointer
		DW	CTOP			;CP	 *****
		DW	NTOP			;NP	 *****
		DW	LASTN			;LAST
                DW      40 DUP(0)               ;frtable
                DW      20 DUP(0)               ;ltable
                DW      0                       ;AMODE
                DW      0                       ;AUTOINC
                DW      0                       ;SDMODE
                DW      0                       ;MIOP
ULAST:

ORG	CODEE					;beginning of the code dictionary

;; Device dependent I/O

;   ?RX 	( -- c )
;		Read input character from RS232 I/O .

		$CODE	3,'?RX',QRX
		DB	01H,22H 		;CLR BX
		DB	0B3H,01H,01H,0E0H,20H	;LDB AL,CTLPORT[0]
		DB	71H,02H,20H		;ANDB AL,#02H
		DB	0DFH,0BH		;JE QRX1
		DB	0B3H,01H,00H,0E0H,20H	;LDB AL,DATPORT[0]
		DB	11H,21H 		;CLRB AH
		DB	0C8H,20H		;PUSH AX
		DB	02H,22H 		;NOT BX
QRX1:		DB	0C8H,22H		;PUSH BX
		$NEXT

;   TX! 	( c -- )
;		Send character c to the output device.

		$CODE	3,'TX!',TXSTO
                DB      0B1H,02H,23H            ;LDB BH,#02H
		DB	0EFH			;LCALL DELAY
		DW	DELAY-$-2
TXSTO1: 	DB	0B3H,01H,01H,0E0H,20H	;LDB AL,CTLPORT[0]
		DB	71H,01H,20H		;ANDB AL,#01H
		DB	0DFH,0F6H		;JE TXSTO1
		DB	0CCH,20H		;POP AX
		DB	0C7H,01H,00H,0E0H,20H	;STB AL,DATPORT[0]
		$NEXT

;   !IO 	( -- )
;		Initialize the serial I/O devices.

		$CODE	3,'!IO',STOIO
		DB	0B1H,40H,20H		;LDB AL,#40H
		DB	0C7H,01H,01H,0E0H,20H	;STB AL,CTLPORT[0]
		DB	0FDH,0FDH		;NOP NOP
		DB	0B1H,4EH,20H		;LDB AL,#4EH
		DB	0C7H,01H,01H,0E0H,20H	;STB AL,CTLPORT[0]
		DB	0B1H,15H,20H		;LDB AL,#15H
		DB	0C7H,01H,01H,0E0H,20H	;STB AL,CTLPORT[0]
		$NEXT


;;		Delay subroutine.

;		delay  approx.x ms.
;		Entrance:BH -------> delay factor.
;		Exit:	 None
;		Usege:BX

DELAY:		DB	0B1H,00H,22H		;LDB BL,#00H
DELAY1: 	DB	0FDH			;NOP
		DB	0E0H,22H,0FCH		;DJNZ BL,DEALY1
		DB	0E0H,23H,0F6H		;DJNZ BH,DELAY
		DB	0F0H			;RET
;

;;; The kernel

;; The inner interpreter

;   doLIT	( -- w )
;		Push an inline literal.

		$CODE	COMPO+5,'doLIT',DOLIT
		DB	0CAH,1BH		;PUSH [IP]+
		$NEXT

;   doLIST	( addr. -- )
;		Process colon list.

		$CODE	COMPO+6,'doLIST',DOLST
		DB	05H,1CH 		;DEC RP
		DB	05H,1CH 		;DEC RP
		DB	0C2H,1CH,1AH		;ST IP,[RP]
		DB	0CCH,1AH		;POP IP
		$NEXT

;   doVAR	( addr. -- addr.)
;		Run time routine for CREATE.

		$CODE	COMPO+5,'doVAR',DOVAR
		DB	0CCH,22H		;POP BX
		DB	65H,02H,00H,22H 	;ADD BX,#02H
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   rdoVAR	( addr. -- addr.)
;		Run time routine for VARIABLE and RCREATE.

		$CODE	COMPO+6,'rdoVAR',RDOVAR
		DB	0CCH,22H		;POP  BX
		DB	0CAH,22H		;PUSH [BX]
		$NEXT

;   doUSER	( addr. -- addr.)
;		Run time routine for user variables.

		$CODE	COMPO+6,'doUSER',DOUSE
		DB	0CCH,22H		;POP  BX
		DB	46H,22H,1EH,20H 	;ADD AX,UP,[BX]
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   doCONST	( addr.--u )
;		Run time code for CONSTANT.

		$CODE	COMPO+7,'doCONST',DOCONST
		DB	0CCH,22H		;POP BX
		DB	0CAH,22H		;PUSH [BX]
		$NEXT

;   doDOES	( -- addr. )
;		Run time routine for DOES>.

		$CODE	COMPO+6,'doDOES',DODOES
		DB	05H,1CH 		;DEC RP
		DB	05H,1CH 		;DEC RP
		DB	0C2H,1CH,1AH		;ST IP,[RP]
		DB	0CCH,22H		;POP BX
		DB	0A2H,23H,1AH		;LD IP,[BX]+
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   next	( -- )
;		Run time code for the single index loop.
;		: next ( -- ) \ hilevel model
;		  r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

		$CODE	COMPO+4,'next',DONXT
		DB	0A2H,1CH,20H		;LD AX,[RP]
		DB	05H,20H 		;DEC AX
		DB	0C2H,1CH,20H		;ST AX,[RP]
		DB	0D3H,08H		;JNC NEXT1
		DB	0A2H,1AH,1AH		;LD IP,[IP]
		$NEXT
NEXT1:		DB	07H,1CH 		;INC RP
		DB	07H,1CH 		;INC RP
		DB	07H,1AH 		;INC IP
		DB	07H,1AH 		;INC IP
		$NEXT

;   ?branch	( f -- )
;		Branch if flag is zero.

		$CODE	COMPO+7,'?branch',QBRAN
		DB	0CCH,22H		;POP BX
		DB	80H,22H,22H		;OR  BX,BX
		DB	0DFH,09H		;JZ  BRAN1
		DB	07H,1AH 		;INC IP
		DB	07H,1AH 		;INC IP
		$NEXT
BRAN1:		DB	0A2H,1AH,1AH		;LD IP,[IP]
		$NEXT

;   branch	( -- )
;		Branch to an inline address.

		$CODE	COMPO+6,'branch',BRAN
		DB	0A2H,1AH,1AH		;LD IP,[IP]
		$NEXT

;   (do)	( limit,index -- )
;		Run time code for DO ... LOOP or DO ... +LOOP structure.

		$CODE	COMPO+4,'(do)',PARDO
		DB	0CCH,24H		;POP CX
		DB	0CCH,22H		;POP BX
		DB	05H,1CH 		;DEC RP
		DB	05H,1CH 		;DEC RP
		DB	0C2H,1CH,22H		;ST BX,[RP]
		DB	05H,1CH 		;DEC RP
		DB	05H,1CH 		;DEC RP
		DB	0C2H,1CH,24H		;ST CX,[RP]
		$NEXT

;   (loop)	( -- )
;		Run time code for DO ... LOOP structure.

		$CODE	COMPO+6,'(loop)',PLOP
		DB	0A1H,01H,00H,24H	;LD CX,#1
PLOP1:		DB	46H,1CH,24H,20H 	;ADD AX,CX,[RP]     AX--Index
		DB	0C2H,1CH,20H		;ST AX,[RP]
		DB	0A3H,1CH,02,22H 	;LD BX,2[RP]	    BX--Limit
		DB	06H,24H 		;EXT CX 	    CX--Increment
		DB	60H,26H,26H		;AND DX,DX
		DB	0DFH,05H		;JZ PLOP2	    If positive,jump
		DB	88H,20H,22H		;CMP BX,AX	    Limit-Index
		DB	20H,03H 		;SJMP PLOP3
PLOP2:		DB	88H,22H,20H		;CMP AX,BX	    Index-Limit
PLOP3:		DB	0D6H,08H		;JGE PLOP4
						;     when increment>0 ,if index>=limit
						;     when increment<0 ,if limit>=index
						;			then exit loop.
		DB	0A2H,1AH,1AH		;LD IP,[IP]
		$NEXT
PLOP4:		DB	65H,04H,00H,1CH 	;ADD RP,#04H	    Pass index and limit
		DB	65H,02H,00H,1AH 	;ADD IP,#02H	    Exit loop
		$NEXT

;   (+loop)	( -- )
;		Run time code for DO ... +LOOP structure.

		$CODE	COMPO+7,'(+loop)',PPLOP
		DB	0CCH,24H		;POP CX
		DB	27H			;SJMP PLOP1
		DB	PLOP1-$-1

;   I		( -- Index value )
;		Copy the inner DO ... LOOP Index onto tops.

		$CODE	1,'I',I
		DB	0CAH,1CH		;PUSH [RP]
		$NEXT

;   J		( -- Index value )
;		Copy the outer DO ... LOOP Index onto tops.

		$CODE	1,'J',J
		DB	0CBH,1CH,04H		;PUSH 4[RP]
		$NEXT

;   (leave)	( -- )
;		Force to exit DO ... LOOP .

		$CODE	COMPO+7,'(leave)',PLEAVE
		DB	65H,04H,00H,1CH 	;ADD RP,#4
		DB	65H,02H,00H,1AH 	;ADD IP,#2
		$NEXT

;   EXECUTE	( ca -- )
;		Execute the word at ca.

		$CODE	7,'EXECUTE',EXECU
		DB	0CCH,22H		;POP BX
		DB	0E3H,22H		;BR  [BX]

;   EXIT	( -- )
;		Terminate a colon definition.

		$CODE	4,'EXIT',EXIT
		DB	0A2H,1DH,1AH		;LD IP,[RP]+
		$NEXT

;   WAIT        ( n -- )
;               Wait for n ms.

                $CODE   4,'WAIT',WAIT
                DB      0CCH,24H                ;POP CX
WAIT1:          DB      0B1H,01H,23H            ;LDB BH,#01H
                DB      0EFH                    ;LCALL DELAY
                DW      DELAY-$-2
                DB      05H,24H                 ;DEC CX
                DB      0D7H,0F6H               ;JNE WAIT1
                $NEXT

;; The stack words

;   !		( w addr. -- )
;		Pop the data stack to memory.

		$CODE	1,'!',STORE
                DB      0CCH,22H                ;POP BX
                DB      0CCH,20H                ;POP AX
                DB      0C6H,23H,20H            ;STB AL,[BX]+
                DB      0C6H,22H,21H            ;STB AH,[BX]
		$NEXT

;   @		( addr. -- w )
;		Push memory location to the data stack.

		$CODE	1,'@',AT
		DB	0CCH,22H		;POP BX
                DB      0B2H,23H,20H            ;LDB AL,[BX]+
                DB      0B2H,22H,21H            ;LDB AH,[BX]
                DB      0C8H,20H                ;PUSH AX
		$NEXT

;   C!		( c b -- )
;		Pop the data stack to byte memory.

		$CODE	2,'C!',CSTOR
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	0C6H,22H,20H		;STB AL,[BX]
		$NEXT

;   C@		( b -- c )
;		Push byte memory location to the data stack.

		$CODE	2,'C@',CAT
		DB	0CCH,22H		;POP BX
		DB	84H,20H,20H		;XOR AX,AX
		DB	0B2H,22H,20H		;LDB AL,[BX]
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   RP@ 	( -- addr.)
;		Push the current RP to the data stack.

		$CODE	3,'RP@',RPAT
		DB	0C8H,1CH		;PUSH RP
		$NEXT

;   RP! 	( addr. -- )
;		Set the return stack pointer.

		$CODE	COMPO+3,'RP!',RPSTO
		DB	0CCH,1CH		;POP RP
		$NEXT

;   R>		( -- w )
;		Pop the return stack to the data stack.

		$CODE	COMPO+2,'R>',RFROM
		DB	0CAH,1DH		;PUSH [RP]+
		$NEXT

;   R@		( -- w )
;		Copy top of return stack to the data stack.

		$CODE	2,'R@',RAT
		DB	0CAH,1CH		;PUSH [RP]
		$NEXT

;   >R		( w -- )
;		Push the data stack to the return stack.

		$CODE	COMPO+2,'>R',TOR
		DB	05H,1CH 		;DEC RP
		DB	05H,1CH 		;DEC RP
		DB	0CEH,1CH		;POP [RP]
		$NEXT

;   SP@ 	( -- addr.)
;		Push the current data stack pointer.

		$CODE	3,'SP@',SPAT
		DB	0C8H,18H		;PUSH SP
		$NEXT

;   SP! 	( addr. -- )
;		Set the data stack pointer.

		$CODE	3,'SP!',SPSTO
		DB	0CCH,18H		;POP SP
		$NEXT

;   DROP	( w -- )
;		Discard top stack item.

		$CODE	4,'DROP',DROP
		DB	07H,18H 		;INC SP
		DB	07H,18H 		;INC SP
		$NEXT

;   DUP 	( w -- w w )
;		Duplicate the top stack item.

		$CODE	3,'DUP',DUPP
		DB	0CAH,18H		;PUSH [SP]
		$NEXT

;   ?DUP	( w -- w w | 0 )
;		Dup tos if its is not zero.

		$CODE	4,'?DUP',QDUP
		DB	0A2H,18H,20H		;LD AX,[SP]
		DB	80H,20H,20H		;OR AX,AX
		DB	0DFH,02H		;JE QDUP1
		DB	0C8H,20H		;PUSH AX
QDUP1:		$NEXT

;   SWAP	( w1 w2 -- w2 w1 )
;		Exchange top two stack items.

		$CODE	4,'SWAP',SWAP
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	0C8H,22H		;PUSH BX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   OVER	( w1 w2 -- w1 w2 w1 )
;		Copy second stack item to top.

		$CODE	4,'OVER',OVER
		DB	0CBH,18H,02H		;PUSH 2[SP]
		$NEXT

;   ROT 	( w1 w2 w3 -- w2 w3 w1 )
;		Rot 3rd item to top.

		$CODE	3,'ROT',ROT
		DB	0CCH,20H		;POP  AX
		DB	0CCH,22H		;POP  BX
		DB	0CCH,24H		;POP  CX
		DB	0C8H,22H		;PUSH BX
		DB	0C8H,20H		;PUSH AX
		DB	0C8H,24H		;PUSH CX
		$NEXT

;   2SWAP       ( d1 -- d2 )
;               Exchange two double integer of tos

                $CODE   5,'2SWAP',DSWAP
                DB      0CCH,20H                ; POP AX
                DB      0CCH,22H                ; POP BX
                DB      0CCH,24H                ; POP CX
                DB      0CCH,26H                ; POP DX
                DB      0C8H,22H                ; PUSH BX
                DB      0C8H,20H                ; PUSH AX
                DB      0C8H,26H                ; PUSH DX
                DB      0C8H,24H                ; PUSH CX
                $NEXT

;   2DROP	( w w -- )
;		Discard two items on stack.

		$CODE	5,'2DROP',DDROP
		DB	65H,04H,00H,18H 	;ADD SP,#4
		$NEXT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;		Duplicate top two items.

		$CODE	4,'2DUP',DDUP
		DB	0CBH,18H,02H		;PUSH 2[SP]
		DB	0CBH,18H,02H		;PUSH 2[SP]
		$NEXT

;   PICK	( ... +n -- ... w )
;		Copy the nth stack item to tos.

		$CODE	4,'PICK',PICK
		DB	0CCH,20H		;POP AX
		DB	09H,01H,20H		;SHL AX,#1
		DB	64H,18H,20H		;ADD AX,SP
		DB	0CAH,20H		;PUSH [AX]
		$NEXT

;   SPLIT       ( w -- wh wl )
;               Split w into wh,wl.

                $CODE   5,'SPLIT',SPLIT
                DB      11H,21H                 ;CLRB AH
                DB      0CCH,22H                ;POP  BX
                DB      0B0H,23H,20H            ;LDB  AL,BH
                DB      0C8H,20H                ;PUSH AX
                DB      0B0H,22H,20H            ;LDB  AL,BL
                DB      0C8H,20H                ;PUSH AX
                $NEXT

;   NIB         ( w -- w)
;               Swap nibble within tos.

                $CODE   3,'NIB',NIBB
                DB      0CCH,22H                ;POP  BX
                DB      0B0H,22H,21H            ;LDB  AH,BL
                DB      0B0H,23H,20H            ;LDB  AL,BH
                DB      0C8H,20H                ;PUSH AX
                $NEXT

;  COMBIN       ( wh wl -- w )
;               Combine two tos into w.

                $CODE   6,'COMBIN',COMBIN
                DB      0CCH,20H                ;POP  AX
                DB      0CCH,22H                ;POP  BX
                DB      0B0H,22H,21H            ;LDB  AH,BL
                DB      0C8H,20H                ;PUSH AX
                $NEXT

;; The compare words

;   0<		( n -- t )
;		Return true if n is negative.

		$CODE	2,'0<',ZLESS
		DB	0CCH,20H		;POP AX
		DB	06H,20H 		;EXT AX
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   0>		( n -- t )
;		Return true if n is positive.

		$CODE	2,'0>',ZGREAT
		DB	0CCH,20H		;POP AX
		DB	89H,00H,00H,20H 	;CMP AX,#0
		DB	0D2H,08H		;JGT ZGREAT1
		DB	0C9H,00H,00H		;PUSH #0
		$NEXT
ZGREAT1:	DB	0C9H,0FFH,0FFH		;PUSH #0FFFFH
		$NEXT

;   0<> 	( n -- t )
;		Return true if n is not zero.

		$CODE	3,'0<>',NOTZERO
		DB	0CCH,20H		;POP AX
		DB	03H,20H 		;NEG AX
		DB	0A8H,20H,20H		;SUBC AX,AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   0=		( w -- t )
;		Return true if w = 0

		$CODE  2,'0=',EQUAL_ZERO
		DB     0CCH,20H 		;POP  AX
		DB     69H,01H,00H,20H		;SUB  AX,#1
		DB     0A8H,20H,20H		;SUBC AX,AX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   =		( w w -- t )
;		Return true if top two are equal.
		$CODE  1,'=',EQUAL
		DB     0CCH,20H 		;POP AX
		DB     0CCH,24H 		;POP CX
		DB     68H,24H,20H		;SUB AX,CX
		DB     69H,01H,00H,20H		;SUB AX,1
		DB     0A8H,20H,20H		;SUBC AX,AX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   <		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE  1,'<',LESS
		DB     0CCH,20H 		;POP AX
		DB     0CCH,22H 		;POP BX
		DB     88H,20H,22H		;CMP BX,AX
		DB     0DEH,0AH 		;JLT LESS1
		DB     68H,20H,20H		;SUB AX,AX
		DB     0C8H,20H 		;PUSH AX
		$NEXT
LESS1:		DB     0A1H,0FFH,0FFH,20H	;LD AX,#0FFFFH
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   >		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE  1,'>',GREAT
		DB     0CCH,20H 		;POP AX
		DB     0CCH,22H 		;POP BX
		DB     88H,20H,22H		;CMP BX,AX
		DB     0D2H,0AH 		;JGT GREAT1
		DB     68H,20H,20H		;SUB AX,AX
		DB     0C8H,20H 		;PUSH AX
		$NEXT
GREAT1: 	DB     0A1H,0FFH,0FFH,20H	;LD AX,#0FFFFH
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   <>		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE  2,'<>',NEQUAL
		DB     01H,24H			;CLR CX
		DB     0CCH,20H 		;POP AX
		DB     0CCH,22H 		;POP BX
		DB     88H,20H,22H		;CMP BX,AX
		DB     0DFH,02H 		;JE  NEQUAL1
		DB     02H,24H			;NOT CX
NEQUAL1:	DB     0C8H,24H 		;PUSH CX
		$NEXT

;   U<		( u u -- t )
;		Unsigned compare of top two items.

		$CODE  2,'U<',ULESS
		DB     0CCH,24H 		;POP CX
		DB     0CCH,20H 		;POP AX
		DB     68H,24H,20H		;SUB AX,CX
		DB     0A8H,20H,20H		;SUBC AX,AX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   U>		( u u -- t )
;		Unsigned compare of top two items.

		$CODE  2,'U>',UGREAT
		DB     0CCH,20H 		;POP AX
		DB     0CCH,24H 		;POP CX
		DB     68H,24H,20H		;SUB AX,CX
		DB     0A8H,20H,20H		;SUBC AX,AX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   MAX 	( n n -- n )
;		Return the greater of two top stack items.

		$CODE  3,'MAX',MAX
		DB	0CCH,20H		;POP AX
		DB	0CCH,22H		;POP BX
		DB	88H,20H,22H		;CMP BX,AX
		DB	0D2H,07H		;JGT MAX1
		DB	0C8H,20H		;PUSH AX
		$NEXT
MAX1:		DB	0C8H,22H		;PUSH BX
		$NEXT

;   MIN 	( n n -- n )
;		Return the smaller of top two stack items.

		$CODE  3,'MIN',MIN
		DB     0CCH,20H 		;POP AX
		DB     0CCH,22H 		;POP BX
		DB     88H,20H,22H		;CMP BX,AX
		DB     0DEH,07H 		;JLT MIN1
		DB     0C8H,20H 		;PUSH AX
		$NEXT
MIN1:		DB     0C8H,22H 		;PUSH BX
		$NEXT

;; The arithmetic and logical words

;   AND 	( w w -- w )
;		Bitwise AND.

		$CODE	3,'AND',ANDD
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	60H,20H,22H		;AND BX,AX
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   OR		( w w -- w )
;		Bitwise inclusive OR.

		$CODE	2,'OR',ORR
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	80H,20H,22H		;OR BX,AX
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   XOR 	( w w -- w )
;		Bitwise exclusive OR.

		$CODE	3,'XOR',XORR
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	84H,20H,22H		;XOR BX,AX
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   NOT 	( w -- w )
;		One's complement of tos.

		$CODE	3,'NOT',INVER
		DB	0CCH,20H		;POP  AX
		DB	02H,20H 		;NOT  AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   NEGATE	( n -- -n )
;		Two's complement of tos.

		$CODE	6,'NEGATE',NEGAT
		DB	0CCH,20H		;POP  AX
		DB	03H,20H 		;NEG  AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   DNEGATE	( d -- -d )
;		Two's complement of top double.

		$CODE	7,'DNEGATE',DNEGA
		DB	0CCH,20H		;POP  AX
		DB	0CCH,22H		;POP  BX
		DB	03H,20H 		;NEG  AX
		DB	03H,22H 		;NEG  BX
		DB	0A9H,00H,00H,20H	;SUBC AX,#0
		DB	0C8H,22H		;PUSH BX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   UM+ 	( u u -- udsum )
;		Add two unsigned single numbers and return a double sum.

		$CODE	3,'UM+',UPLUS
		DB	84H,24H,24H		;XOR CX,CX
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	64H,22H,20H		;ADD AX,BX
		DB	0A5H,00H,00H,24H	;ADDC CX,0
		DB	0C8H,20H		;PUSH AX
		DB	0C8H,24H		;PUSH CX
		$NEXT

;   +		( w w -- sum )
;		Add top two items.

		$CODE	1,'+',PLUS
		DB	0CCH,22H		;POP  BX
		DB	0CCH,20H		;POP  AX
		DB	64H,22H,20H		;ADD  AX,BX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   -		( n1 n2 -- n1-n2 )
;		Subtraction.

		$CODE  1,'-',SUBB
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	68H,22H,20H		;SUB AX,BX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;; Divide

;   UM/MOD	( udl udh un -- ur uq )
;		Unsigned divide of a double by a single. Return mod and quotient.

		$CODE  6,'UM/MOD',UMMOD
		DB     0CCH,24H 		;POP CX
		DB     0CCH,22H 		;POP BX
		DB     0CCH,20H 		;POP AX
                DB     8CH,24H,20H              ;DIVU AX,CX
		DB     0C8H,22H 		;PUSH BX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   M/MOD	( d n -- r q )
;		Signed floored divide of double by single. Return mod and quotient.

		$CODE  5,'M/MOD',MSMOD
		DB     0CCH,24H 		;POP CX
		DB     0CCH,22H 		;POP BX
		DB     0CCH,20H 		;POP AX
		DB     0FEH,8CH,24H,20H 	;DIV AX,CX
		DB     0C8H,22H 		;PUSH BX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   /MOD	( n n -- r q )
;		Signed divide. Return mod and quotient.

		$CODE	4,'/MOD',SLMOD
		DB	0CCH,24H		;POP CX
		DB	0CCH,20H		;POP AX
                DB      01H,22H                 ;CLR BX
                DB      8CH,24H,20H             ;DIVU AX,CX
                DB      16H,20H                 ;EXTB AX
		DB	0C8H,22H		;PUSH BX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   MOD 	( n n -- r )
;		Signed divide. Return mod only.

		$CODE	3,'MOD',MODD
		DB	0CCH,24H		;POP CX
		DB	0CCH,20H		;POP AX
		DB	06H,20H 		;EXT AX
		DB	0FEH,8CH,24H,20H	;DIV AX,CX
		DB	0C8H,22H		;PUSH BX
		$NEXT

;   /		( n n -- q )
;		Signed divide. Return quotient only.

		$CODE	1,'/',SLASH
		DB	0CCH,24H		;POP CX
		DB	0CCH,20H		;POP AX
		DB	06H,20H 		;EXT AX
		DB	0FEH,8CH,24H,20H	;DIV AX,CX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;; Multiply

;   UM* 	( u u -- ud )
;		Unsigned multiply. Return double product.

		$CODE  3,'UM*',UMSTA
		DB     0CCH,24H 		;POP CX
		DB     0CCH,20H 		;POP AX
		DB     6CH,24H,20H		;MULU AX,CX
		DB     0C8H,20H 		;PUSH AX
		DB     0C8H,22H 		;PUSH BX
		$NEXT

;   *		( n n -- n )
;		Signed multiply. Return single product.

		$CODE	1,'*',STAR
		DB     0CCH,24H 		;POP CX
		DB     0CCH,20H 		;POP AX
		DB     6CH,24H,20H		;MULU AX,CX
		DB     0C8H,20H 		;PUSH AX
		$NEXT

;   M*		( n n -- d )
;		Signed multiply. Return double product.

		$CODE  2,'M*',MSTAR
		DB     0CCH,24H 		;POP CX
		DB     0CCH,20H 		;POP AX
		DB     0FEH,6CH,24H,20H 	;MUL AX,CX
		DB     0C8H,20H 		;PUSH AX
		DB     0C8H,22H 		;PUSH BX
		$NEXT

;   */MOD	( n1 n2 n3 -- r q )
;		Multiply n1 and n2, then divide by n3. Return mod and quotient.

		$COLON	5,'*/MOD',SSMOD
		DW	TOR,MSTAR,RFROM,MSMOD,EXIT


;   */		( n1 n2 n3 -- q )
;		Multiply n1 by n2, then divide by n3. Return quotient only.

		$COLON	2,'*/',STASL
		DW	SSMOD,SWAP,DROP,EXIT

;; Miscellaneous

;   1+		( addr. -- addr.+1 )
;		Increment by 1.

		$CODE	2,'1+',INCR
		DB	0CCH,20H		;POP AX
		DB	07H,20H 		;INC AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   1-		( addr. -- addr.-1 )
;		Decrement by 1.

		$CODE	2,'1-',DECR
		DB	0CCH,20H		;POP AX
		DB	05H,20H 		;DEC AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   2+		( addr. -- addr.)
;		Add cell size in byte to address.

		$CODE	2,'2+',CELLP
		DB	0CCH,20H		;POP AX
		DB	07H,20H 		;INC AX
		DB	07H,20H 		;INC AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   2-		( addr. -- addr.)
;		Subtract cell size in byte from address.

		$CODE	2,'2-',CELLM
		DB	0CCH,20H		;POP AX
		DB	05H,20H 		;DEC AX
		DB	05H,20H 		;DEC AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   2*		( n -- n )
;		Multiply tos by cell size in bytes.

		$CODE	2,'2*',CELLS
		DB	0CCH,20H		;POP AX
		DB	64H,20H,20H		;ADD AX,AX
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   2/		( n -- n )
;		Divide tos by cell size in bytes.

		$CODE	2,'2/',CELLD
		DB	0CCH,20H		;POP AX
		DB	0AH,01H,20H		;SHRA AX,#1
		DB	0C8H,20H		;PUSH AX
		$NEXT

;   2MOD	( n -- r )
;		Mod tos by 2.

		$CODE	4,'2MOD',MOD2
		DB	84H,24H,24H		;XOR CX,CX
		DB	0CCH,20H		;POP AX
		DB	08H,01H,20H		;SHR AX,#1
		DB	0A4H,24H,24H		;ADDC CX,CX
		DB	0C8H,24H		;PUSH CX
		$NEXT

;   DI		( -- )
;		Interrupt disable.

		$CODE	2,'DI',DII
		DB	0FAH			;DI
		$NEXT

;   EI		( -- )
;		Interrupt enable.

		$CODE	2,'EI',EII
		DB	0FBH			;EI
		$NEXT

;   RET         ( -- )
;               Subroutine return.

                $CODE   3,'RET',RETT
                DB      0CCH,26H                ;POP DX
                DB      0CCH,24H                ;POP CX
                DB      0CCH,22H                ;POP BX
                DB      0CCH,20H                ;POP AX
                DB      0CCH,1AH                ;POP IP
                DB      0CCH,1CH                ;POP RP
                DB      0F3H                    ;POPF
                DB      0F0H                    ;RET

;   UP		( -- addr.)
;		Pointer to the user area.

		$CODE	2,'UP',UP
		DB	0C8H,1EH		;PUSH UP
		$NEXT

;   VP		( -- addr. )
;		Pointer to the top available RAM.

		$CODE	2,'VP',VP
		DB	0C8H,28H		;PUSH VP
		$NEXT

;; Memory access

;   +!		( n addr. -- )
;		Add n to the contents at address a.

		$CODE	2,'+!',PSTOR
		DB	0CCH,22H		;POP BX
		DB	0CCH,20H		;POP AX
		DB	66H,22H,20H		;ADD AX,[BX]
		DB	0C2H,22H,20H		;ST AX,[BX]
		$NEXT

;   2!		( d addr. -- )
;		Store the double integer to address a.

		$CODE	2,'2!',DSTOR
		DB	0CCH,22H		;POP BX
		DB	0CEH,22H		;POP [BX]
		DB	0CFH,22H,02H		;POP 2[BX]
		$NEXT

;   2@		( addr. -- d )
;		Fetch double integer from address a.

		$CODE	2,'2@',DAT
		DB	0CCH,22H		;POP BX
		DB	0CBH,22H,02H		;PUSH 2[BX]
		DB	0CAH,22H		;PUSH [BX]
		$NEXT

;   CMOVE	( b1 b2 u -- )
;		Copy u bytes from b1 to b2.

		$CODE	5,'CMOVE',CMOVE
		DB	0CCH,24H		;POP CX
		DB	0CCH,26H		;POP DX
		DB	0CCH,22H		;POP BX
		DB	88H,00H,24H		;CMP CX,R0
		DB	0DFH,0AH		;JE CMOV2
CMOV1:		DB	0B2H,23H,20H		;LDB AL,[BX]+
		DB	0C6H,27H,20H		;STB AL,[DX]+
		DB	05H,24H 		;DEC CX
		DB	0D7H,0F6H		;JNE CMOV1
CMOV2:		$NEXT

;   CMOVE>	( b1 b2 u -- )
;		Copy u bytes from (b1+u-1) downto (b2+u-1).

		$CODE	6,'CMOVE>',CMVET
		DB	0CCH,24H		;POP CX
		DB	0CCH,26H		;POP DX
		DB	0CCH,22H		;POP BX
		DB	88H,00H,24H		;CMP CX,R0
		DB	0DFH,14H		;JE CMVET2
		DB	064H,24H,22H		;ADD BX,CX
		DB	064H,24H,26H		;ADD DX,CX
CMVET1: 	DB	05H,22H 		;DEC BX
		DB	05H,26H 		;DEC DX
		DB	0B2H,22H,20H		;LDB AL,[BX]
		DB	0C6H,26H,20H		;STB AL,[DX]
		DB	05H,24H 		;DEC CX
		DB	0D7H,0F2H		;JNE CMVET1
CMVET2: 	$NEXT

;   FILL	( b u c -- )
;		Fill u bytes of character c to area beginning at b.

		$CODE	4,'FILL',FILL
		DB	0CCH,20H		;POP AX
		DB	0CCH,24H		;POP CX
		DB	0CCH,22H		;POP BX
		DB	88H,00H,24H		;CMP BX,R0
		DB	0DFH,07H		;JE FILL2
FILL1:		DB	0C6H,23H,20H		;STB AL,[BX]+
		DB	05H,24H 		;DEC CX
		DB	0D7H,0F9H		;JNE FILL1
FILL2:		$NEXT

;; System and user variables


;   SP0 	( -- addr.)
;		Pointer to bottom of the data stack.

		$USER	3,'SP0',SZERO

;   RP0 	( -- addr.)
;		Pointer to bottom of the return stack.

		$USER	3,'RP0',RZERO

;   UP0         ( -- addr.)
;               Pointer to bottom of the user variable area.

                $USER   3,'UP0',UPZERO

;   '?KEY       ( -- addr.)
;		Execution vector of ?KEY.

		$USER	5,"'?KEY",TQKEY

;   'EMIT       ( -- addr.)
;		Execution vector of EMIT.

		$USER	5,"'EMIT",TEMIT

;   'EXPECT     ( -- addr.)
;		Execution vector of EXPECT.

		$USER	7,"'EXPECT",TEXPE

;   'TAP        ( -- addr.)
;		Execution vector of TAP.

		$USER	4,"'TAP",TTAP

;   'ECHO       ( -- addr.)
;		Execution vector of ECHO.

		$USER	5,"'ECHO",TECHO

;   'PROMPT     ( -- addr.)
;		Execution vector of PROMPT.

		$USER	7,"'PROMPT",TPROM

;   BASE	( -- addr.)
;		Storage of the radix base for numeric I/O.

		$USER	4,'BASE',BASE

;   tmp 	( -- addr.)
;		A temporary storage location used in parse and find.

		$USER	COMPO+3,'tmp',TEMP

;   SPAN	( -- addr.)
;		Hold character count received by EXPECT.

		$USER	4,'SPAN',SPAN

;   >IN 	( -- addr.)
;		Hold the character pointer while parsing input stream.

		$USER	3,'>IN',INN

;   #TIB	( -- addr.)
;		Hold the current count in and address of the terminal input buffer.

		$USER	4,'#TIB',NTIB
		_USER = _USER+CELLL

;   CSP 	( -- addr.)
;		Hold the stack pointer for error checking.

		$USER	3,'CSP',CSP

;   'EVAL       ( -- addr.)
;		Execution vector of EVAL.

		$USER	5,"'EVAL",TEVAL

;   'NUMBER     ( -- addr.)
;		Execution vector of NUMBER?.

		$USER	7,"'NUMBER",TNUMB

;   HLD 	( -- addr.)
;		Hold a pointer in building a numeric output string.

		$USER	3,'HLD',HLD

;   HANDLER	( -- addr.)
;		Hold the return stack pointer for error handling.

		$USER	7,'HANDLER',HANDL

;   VOC-LINK	( -- addr. )
;		Hold the newest defined vocabulary.

		$USER	8,'VOC-LINK',VOCLK

;   CONTEXT	( -- addr.)
;		A area to specify vocabulary search order.

		$USER	7,'CONTEXT',CNTXT
		_USER = _USER+(VOCSS+1)*CELLL	;vocabulary stack

;   CURRENT	( -- addr.)
;		Point to the vocabulary to be extended.

		$USER	7,'CURRENT',CRRNT
		_USER = _USER+CELLL		;vocabulary link pointer

;   CP		( -- addr.)
;		Point to the top of the code dictionary.

		$USER	2,'CP',CP

;   NP		( -- addr.)
;		Point to the bottom of the name dictionary.

		$USER	2,'NP',NP

;   LAST	( -- addr.)
;		Point to the last name in the name dictionary.

		$USER	4,'LAST',LAST

;; Common functions

;   doVOC	( addr. -- )
;		Run time action of VOCABULARY's.

		$COLON	COMPO+5,'doVOC',DOVOC
		DW	AT,CNTXT,STORE,EXIT

;   FORTH	( -- )
;		Make FORTH the context vocabulary.

		$DOES	IMEDD+5,'FORTH',FORTH
		DW	VOCA1,VPP0

;   ASSEMBLER	( --)
;		Make ASSEMBLER the context vocabulary.

		$DOES	IMEDD+9,'ASSEMBLER',ASSMB
		DW	VOCA1,VPP0+6

;   0>= 	( n -- t )
;		Return true if n>=0.

		$COLON	3,'0>=',ZGE
		DW	ZLESS,INVER,EXIT

;   0<= 	( n -- t )
;		Return true if n <=0.

		$COLON	3,'0<=',ZLE
		DW	ZGREAT,INVER,EXIT

;   0=< 	( n -- t )
;		Same as 0<=.

		$COLON	3,'0=<',ZLE_ALIAS
		DW	ZLE,EXIT

;   >=		( n1 n2 -- t )
;		Return true if n1>=n2.

		$COLON	2,'>=',GE
		DW	LESS,INVER,EXIT

;   <=		( n1 n2 -- t )
;		Return true if n1<=n2.

		$COLON	2,'<=',LE1
		DW	GREAT,INVER,EXIT

;   =<		( n1,n2 -- t )
;		Same as <=

		$COLON	2,'=<',LE2
		DW	LE1,EXIT

;   ABS 	( n -- n )
;		Return the absolute value of n.

		$COLON	3,'ABS',ABSS
		DW	DUPP,ZLESS
		DW	QBRAN,ABS1
		DW	NEGAT
ABS1:		DW	EXIT

;   (;CODE)	( -- )
;		The run time procedure for ;CODE,

		$COLON	7,'(;CODE)',PSCODE
		DW	RFROM,LAST,AT
		DW	NAMET,STORE,EXIT

;   WITHIN	( u ul uh -- t )
;		Return true if u is within the range of ul and uh. ( ul <= u < uh )

		$COLON	6,'WITHIN',WITHI
		DW	OVER,SUBB,TOR
		DW	SUBB,RFROM,ULESS,EXIT


;   ALIGNED	( b -- addr.)
;		Align address to the cell boundary.

		$COLON	7,'ALIGNED',ALGND
		DW	DUPP
		DW	MOD2
		DW	QBRAN,ALGN1
		DW	INCR
ALGN1:		DW	EXIT

;   BL		( -- 32 )
;		Return 32, the blank character.

		$COLON	2,'BL',BLANK
		DW	DOLIT,' ',EXIT

;   >CHAR	( c -- c )
;		Filter non-printing characters.

		$COLON	5,'>CHAR',TCHAR
		DW	DOLIT,07FH,ANDD,DUPP	;mask msb
		DW	DOLIT,127,BLANK,WITHI	;check for printable
		DW	QBRAN,TCHA1
		DW	DROP,DOLIT,'_'          ;replace non-printables
TCHA1:		DW	EXIT

;   DEPTH	( -- n )
;		Return the depth of the data stack.

		$COLON	5,'DEPTH',DEPTH
		DW	SPAT,SZERO,AT,SWAP,SUBB
		DW	CELLD,EXIT

;   COUNT	( b -- b+1 n )
;		Return count byte of a string and add 1 to byte address.

		$COLON	5,'COUNT',COUNT
		DW	DUPP,INCR
		DW	SWAP,CAT,EXIT

;   ERASE       ( b u -- )
;		Clear u bytes from b.

                $COLON  5,'ERASE',ERASE
		DW	DOLIT,0,FILL,EXIT

;   HERE	( -- addr.)
;		Return the top of the code dictionary.

		$COLON	4,'HERE',HERE
		DW	CP,AT,EXIT


;   RHERE	( -- addr.)
;		Return the top of RAM.

		$CODE	5,'RHERE',RHERE
		DB	0C8H,28H		;PUSH VP
		$NEXT

;   PAD 	( -- addr.)
;               Return the address of the text buffer above the ram variable.

		$COLON	3,'PAD',PAD
                DW      RHERE,DOLIT,80,PLUS,EXIT

;   TIB 	( -- addr.)
;		Return the address of the terminal input buffer.

		$COLON	3,'TIB',TIB
		DW	NTIB,CELLP,AT,EXIT

;   @EXECUTE	( addr. -- )
;		Execute vector stored in address a.

		$COLON	8,'@EXECUTE',ATEXE
		DW	AT,QDUP 		;?address or zero
		DW	QBRAN,EXE1
		DW	EXECU			;execute if non-zero
EXE1:		DW	EXIT			;do nothing if zero

;   -TRAILING	( b u -- b u )
;		Adjust the count to eliminate trailing white space.

		$COLON	9,'-TRAILING',DTRAI
		DW	TOR
		DW	BRAN,DTRA2
DTRA1:		DW	BLANK,OVER,RAT,PLUS,CAT,LESS
		DW	QBRAN,DTRA2
		DW	RFROM,INCR,EXIT
DTRA2:		DW	DONXT,DTRA1
		DW	DOLIT,0,EXIT

;   PACK$	( b u addr. -- addr.)
;		Build a counted string with u characters from b. Null fill.

		$COLON	5,'PACK$',PACKS
		DW	ALGND,DUPP,TOR		;strings only on cell boundary
		DW	OVER,DUPP
		DW	MOD2			;count mod cell
		DW	SUBB,OVER,PLUS
		DW	DOLIT,0,SWAP,STORE	;null fill cell
		DW	DDUP,CSTOR,INCR 	;save count
		DW	SWAP,CMOVE,RFROM,EXIT	;move string

;; Numeric output, single precision

;   DIGIT	( u -- c )
;		Convert digit u to a character.

		$COLON	5,'DIGIT',DIGIT
		DW	DOLIT,9,OVER,LESS
		DW	DOLIT,7,ANDD,PLUS
		DW	DOLIT,'0',PLUS,EXIT

;   EXTRACT	( n base -- n c )
;		Extract the least significant digit from n.

		$COLON	7,'EXTRACT',EXTRC
		DW	DOLIT,0,SWAP,UMMOD
		DW	SWAP,DIGIT,EXIT

;   <#		( -- )
;		Initiate the numeric output process.

		$COLON	2,'<#',BDIGS
		DW	PAD,HLD,STORE,EXIT

;   HOLD	( c -- )
;		Insert a character into the numeric output string.

		$COLON	4,'HOLD',HOLD
		DW	HLD,AT,DECR
		DW	DUPP,HLD,STORE,CSTOR,EXIT

;   #		( u -- u )
;		Extract one digit from u and append the digit to output string.

		$COLON	1,'#',DIG
		DW	BASE,AT,EXTRC,HOLD,EXIT

;   #S		( u -- 0 )
;		Convert u until all digits are added to the output string.

		$COLON	2,'#S',DIGS
DIGS1:		DW	DIG,DUPP
		DW	QBRAN,DIGS2
		DW	BRAN,DIGS1
DIGS2:		DW	EXIT

;   SIGN	( n -- )
;		Add a minus sign to the numeric output string.

		$COLON	4,'SIGN',SIGN
		DW	ZLESS
		DW	QBRAN,SIGN1
		DW	DOLIT,'-',HOLD
SIGN1:		DW	EXIT

;   #>		( w -- b u )
;		Prepare the output string to be TYPE'd.

		$COLON	2,'#>',EDIGS
		DW	DROP,HLD,AT
		DW	PAD,OVER,SUBB,EXIT

;   str 	( w -- b u )
;		Convert a signed integer to a numeric string.

		$COLON	3,'str',STR
		DW	DUPP,TOR,ABSS
		DW	BDIGS,DIGS,RFROM
		DW	SIGN,EDIGS,EXIT

;   HEX 	( -- )
;		Use radix 16 as base for numeric conversions.

		$COLON	3,'HEX',HEX
		DW	DOLIT,16,BASE,STORE,EXIT


;   DECIMAL	( -- )
;		Use radix 10 as base for numeric conversions.

		$COLON	7,'DECIMAL',DECIM
		DW	DOLIT,10,BASE,STORE,EXIT

;; Numeric input, single precision

;   DIGIT?	( c base -- u t )
;		Convert a character to its numeric value. A flag indicates success.

		$COLON	6,'DIGIT?',DIGTQ
		DW	TOR,DOLIT,'0',SUBB
		DW	DOLIT,9,OVER,LESS
		DW	QBRAN,DGTQ1
		DW	DOLIT,7,SUBB
		DW	DUPP,DOLIT,10,LESS,ORR
DGTQ1:		DW	DUPP,RFROM,ULESS,EXIT

;   NUMBER?	( addr. -- n T | addr. F )
;		Convert a number string to integer. Push a flag on tos.

		$COLON	7,'NUMBER?',NUMBQ
		DW	BASE,AT,TOR,DOLIT,0,OVER,COUNT
		DW	OVER,CAT,DOLIT,'$',EQUAL
		DW	QBRAN,NUMQ1
		DW	HEX,SWAP,INCR
		DW	SWAP,DECR
NUMQ1:		DW	OVER,CAT,DOLIT,'-',EQUAL,TOR
		DW	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
		DW	QBRAN,NUMQ6
		DW	DECR,TOR
NUMQ2:		DW	DUPP,TOR,CAT,BASE,AT,DIGTQ
		DW	QBRAN,NUMQ4
		DW	SWAP,BASE,AT,STAR,PLUS,RFROM
		DW	INCR
		DW	DONXT,NUMQ2
		DW	RAT,SWAP,DROP
		DW	QBRAN,NUMQ3
		DW	NEGAT
NUMQ3:		DW	SWAP
		DW	BRAN,NUMQ5
NUMQ4:		DW	RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:		DW	DUPP
NUMQ6:		DW	RFROM,DDROP
		DW	RFROM,BASE,STORE,EXIT

;; Basic I/O

;   ?KEY	( -- c T | F)
;		Return input character and true , or a false if no input.

		$COLON	4,'?KEY',QKEY
		DW	TQKEY,ATEXE,EXIT

;   KEY 	( -- c)
;		Wait for and return an input character.

		$COLON	3,'KEY',KEY
KEY1:		DW	QKEY
		DW	QBRAN,KEY1
		DW	EXIT

;   EMIT	( c -- )
;		Send a character to the output device.

		$COLON	4,'EMIT',EMIT
		DW	TEMIT,ATEXE,EXIT

;   PACE	( -- )
;		Send a pace character for the file downloading process.

		$COLON	4,'PACE',PACE
		DW	DOLIT,11,EMIT,EXIT

;   send_over	( -- )
;		Tell host sending process is over.

		$COLON	9,'send_over',S_O
		DW	DOLIT,CTRL_Z,EMIT,EXIT

;   SPACE	( -- )
;		Send the blank character to the output device.

		$COLON	5,'SPACE',SPACE
		DW	BLANK,EMIT,EXIT

;   SPACES	( +n -- )
;		Send n spaces to the output device.

		$COLON	6,'SPACES',SPACS
		DW	DOLIT,0,MAX,TOR
		DW	BRAN,CHAR2
CHAR1:		DW	SPACE
CHAR2:		DW	DONXT,CHAR1
		DW	EXIT

;   TYPE	( b u -- )
;		Output u characters from b.

		$COLON	4,'TYPE',TYPES
		DW	TOR
		DW	BRAN,TYPE2
TYPE1:		DW	DUPP,CAT,EMIT
		DW	INCR
TYPE2:		DW	DONXT,TYPE1
		DW	DROP,EXIT

;   CR		( -- )
;		Output a carriage return and a line feed.

		$COLON	2,'CR',CR
		DW	DOLIT,CRR,EMIT
		DW	DOLIT,LF,EMIT,EXIT

;   do$ 	( -- addr.)
;		Return the address of a compiled string.

		$COLON	COMPO+3,'do$',DOSTR
		DW	RFROM,RAT,RFROM,COUNT,PLUS
		DW	ALGND,TOR,SWAP,TOR,EXIT

;   $"|         ( -- addr.)
;		Run time routine compiled by $". Return address of a compiled string.

		$COLON	COMPO+3,'$"|',STRQP
		DW	DOSTR,EXIT		;force a call to do$

;   ."|         ( -- )
;		Run time routine of ." . Output a compiled string.

		$COLON	COMPO+3,'."|',DOTQP
		DW	DOSTR,COUNT,TYPES,EXIT

;   .R		( n +n -- )
;		Display an integer in a field of n columns, right justified.

		$COLON	2,'.R',DOTR
		DW	TOR,STR,RFROM,OVER,SUBB
		DW	SPACS,TYPES,EXIT

;   U.R 	( u +n -- )
;		Display an unsigned integer in n column, right justified.

		$COLON	3,'U.R',UDOTR
		DW	TOR,BDIGS,DIGS,EDIGS
		DW	RFROM,OVER,SUBB
		DW	SPACS,TYPES,EXIT

;   U.		( u -- )
;		Display an unsigned integer in free format.

		$COLON	2,'U.',UDOT
		DW	BDIGS,DIGS,EDIGS
		DW	SPACE,TYPES,EXIT

;   .		( w -- )
;		Display an integer in free format, preceeded by a space.

		$COLON	1,'.',DOT
		DW	BASE,AT,DOLIT,10,XORR	;?decimal
		DW	QBRAN,DOT1
		DW	UDOT,EXIT		;no, display unsigned
DOT1:		DW	STR,SPACE,TYPES,EXIT	;yes, display signed


;   ?		( addr. -- )
;		Display the contents in a memory cell.

		$COLON	1,'?',QUEST
		DW	AT,DOT,EXIT

;; Parsing

;   parse	( b u c -- b u delta ; <string> )
;		Scan string delimited by c. Return found string and its offset.

		$COLON	5,'parse',PARS
		DW	TEMP,STORE,OVER,TOR,DUPP
		DW	QBRAN,PARS8
		DW	DECR,TEMP,AT,BLANK,EQUAL
		DW	QBRAN,PARS3
		DW	TOR
PARS1:		DW	BLANK,OVER,CAT		;skip leading blanks ONLY
		DW	SUBB,ZLESS,INVER
		DW	QBRAN,PARS2
		DW	INCR
		DW	DONXT,PARS1
		DW	RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:		DW	RFROM
PARS3:		DW	OVER,SWAP
		DW	TOR
PARS4:		DW	TEMP,AT,OVER,CAT,SUBB	;scan for delimiter
		DW	TEMP,AT,BLANK,EQUAL
		DW	QBRAN,PARS5
		DW	ZLESS
PARS5:		DW	QBRAN,PARS6
		DW	INCR
		DW	DONXT,PARS4
		DW	DUPP,TOR
		DW	BRAN,PARS7
PARS6:		DW	RFROM,DROP,DUPP
		DW	INCR,TOR
PARS7:		DW	OVER,SUBB
		DW	RFROM,RFROM,SUBB,EXIT
PARS8:		DW	OVER,RFROM,SUBB,EXIT

;   PARSE	( c -- b u ; <string> )
;		Scan input stream and return counted string delimited by c.

		$COLON	5,'PARSE',PARSE
		DW	TOR,TIB,INN,AT,PLUS	;current input buffer pointer
		DW	NTIB,AT,INN,AT,SUBB	;remaining count
		DW	RFROM,PARS,INN,PSTOR,EXIT

;   .(		( -- )
;		Output following string up to next ) .

		$COLON	IMEDD+2,'.(',DOTPR
		DW	DOLIT,')',PARSE,TYPES,EXIT

;   (		( -- )
;		Ignore following string up to next ) . A comment.

		$COLON	IMEDD+1,'(',PAREN
		DW	DOLIT,')',PARSE,DDROP,EXIT

;   \		( -- )
;		Ignore following text till the end of line.

		$COLON	IMEDD+1,'\',BKSLA
		DW	NTIB,AT,INN,STORE,EXIT

;   CHAR	( -- c )
;		Parse next word and return its first character.

		$COLON	4,'CHAR',CHAR
		DW	BLANK,PARSE,DROP,CAT,EXIT

;   TOKEN	( -- addr. ; <string> )
;		Parse a word from input stream and copy it to name dictionary.

		$COLON	5,'TOKEN',TOKEN
		DW	BLANK,PARSE,DOLIT,31,MIN
		DW	NP,AT,OVER,SUBB,CELLM
		DW	PACKS,EXIT

;   WORD	( c -- addr. ; <string> )
;		Parse a word from input stream and copy it to code dictionary.

		$COLON	4,'WORD',WORDD
		DW	PARSE,HERE,PACKS,EXIT

;; Dictionary search

;   NAME>	( na -- ca )
;		Return a code address given a name address.

		$COLON	5,'NAME>',NAMET
		DW	CELLM,CELLM,AT,EXIT

;   SAME?	( addr.  addr. u -- addr. addr. f \ -0+ )
;		Compare u cells in two strings. Return 0 if identical.

		$COLON	5,'SAME?',SAMEQ
		DW	TOR
		DW	BRAN,SAME2
SAME1:		DW	OVER,RAT,CELLS,PLUS,AT
		DW	OVER,RAT,CELLS,PLUS,AT
		DW	SUBB,QDUP
		DW	QBRAN,SAME2
		DW	RFROM,DROP,EXIT
SAME2:		DW	DONXT,SAME1
		DW	DOLIT,0,EXIT

;   find	( addr. va -- ca na | addr. F )
;		Search a vocabulary for a string. Return ca and na if succeeded.

		$COLON	4,'find',FIND
		DW	SWAP,DUPP,CAT
		DW	CELLD,TEMP,STORE
		DW	DUPP,AT,TOR,CELLP,SWAP
FIND1:		DW	AT,DUPP
		DW	QBRAN,FIND6
		DW	DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
		DW	QBRAN,FIND2
		DW	CELLP,DOLIT,-1
		DW	BRAN,FIND4
FIND2:		DW	CELLP,TEMP,AT,SAMEQ
FIND4:		DW	QBRAN,FIND5
		DW	CELLM,CELLM
		DW	BRAN,FIND1
FIND5:		DW	RFROM,DROP,SWAP,DROP
		DW	CELLM
		DW	DUPP,NAMET,SWAP,EXIT
FIND6:		DW	RFROM,DROP
		DW	SWAP,CELLM,SWAP,EXIT

;   NAME?	( addr. -- ca na | addr. F )
;		Search all context vocabularies for a string.

		$COLON	5,'NAME?',NAMEQ
		DW	CNTXT,DUPP,DAT,XORR
		DW	QBRAN,NAMQ1
		DW	CELLM
NAMQ1:		DW	TOR
NAMQ2:		DW	RFROM,CELLP,DUPP,TOR
		DW	AT,QDUP
		DW	QBRAN,NAMQ3
		DW	FIND,QDUP
		DW	QBRAN,NAMQ2
		DW	RFROM,DROP,EXIT
NAMQ3:		DW	RFROM,DROP
		DW	DOLIT,0,EXIT

;; Terminal response

;   ^H		( bot eot cur -- bot eot cur )
;		Backup the cursor by one character.

		$COLON	2,'^H',BKSP
		DW	TOR,OVER,RFROM,SWAP,OVER,XORR
		DW	QBRAN,BACK1
		DW	DECR
BACK1:		DW	EXIT

;   TAP 	( bot eot cur c -- bot eot cur )
;		Accept the key stroke and bump the cursor.

		$COLON	3,'TAP',TAP
		DW	OVER,CSTOR,INCR,EXIT

;   kTAP	( bot eot cur c -- bot eot cur )
;		Process a key stroke, CR or backspace.

		$COLON	4,'kTAP',KTAP
		DW	DUPP,DOLIT,CRR,XORR
		DW	QBRAN,KTAP2
		DW	DOLIT,BKSPP,XORR
		DW	QBRAN,KTAP1
		DW	BLANK,TAP,EXIT
KTAP1:		DW	BKSP,EXIT
KTAP2:		DW	DROP,SWAP,DROP,DUPP,EXIT

;   accept	( b u -- b u )
;		Accept characters to input buffer. Return with actual count.

		$COLON	6,'accept',ACCEP
		DW	OVER,PLUS,OVER
ACCP1:		DW	DDUP,XORR
		DW	QBRAN,ACCP4
		DW	KEY,DUPP
		DW	BLANK,DOLIT,127,WITHI
		DW	QBRAN,ACCP2
		DW	TAP
		DW	BRAN,ACCP1
ACCP2:		DW	TTAP,ATEXE
ACCP3:		DW	BRAN,ACCP1
ACCP4:		DW	DROP,OVER,SUBB,EXIT

;   EXPECT	( b u -- )
;		Accept input stream and store count in SPAN.

		$COLON	6,'EXPECT',EXPEC
		DW	TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;   QUERY	( -- )
;		Accept input stream to terminal input buffer.

		$COLON	5,'QUERY',QUERY
		DW	TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
		DW	DROP,DOLIT,0,INN,STORE,EXIT

;; Error handling

;   CATCH	( ca -- 0 | err# )
;		Execute word at ca and set up an error frame for it.

		$COLON	5,'CATCH',CATCH
		DW	SPAT,TOR,HANDL,AT,TOR	;save error frame
		DW	RPAT,HANDL,STORE,EXECU	;execute
		DW	RFROM,HANDL,STORE	;restore error frame
		DW	RFROM,DROP,DOLIT,0,EXIT ;no error

;   THROW	( err# -- err# )
;		Reset system to current local error frame an update error flag.

		$COLON	5,'THROW',THROW
		DW	HANDL,AT,RPSTO		;restore return stack
		DW	RFROM,HANDL,STORE	;restore handler frame
		DW	RFROM,SWAP,TOR,SPSTO	;restore data stack
		DW	DROP,RFROM,EXIT

;   NULL$	( -- addr.)
;		Return address of a null string with zero count.

		$VAR	5,'NULL$',NULLS
		DB	0FDH,0FDH		; NOP NOP
		DW	0
		DB	99,111,121,111,116,101

;   ABORT	( -- )
;		Reset data stack and jump to QUIT.

		$COLON	5,'ABORT',ABORT
		DW	NULLS,THROW

;   abort"      ( f -- )
;		Run time routine of ABORT" . Abort with a message.

		$COLON	COMPO+6,'abort"',ABORQ
		DW	QBRAN,ABOR1		;text flag
		DW	DOSTR,THROW		;pass error string
ABOR1:		DW	DOSTR,DROP,EXIT 	;drop error

;; The text interpreter

;   $INTERPRET	( addr. -- )
;		Interpret a word. If failed, try to convert it to an integer.

		$COLON	10,'$INTERPRET',INTER
		DW	NAMEQ,QDUP		;?defined
		DW	QBRAN,INTE1
		DW	AT,DOLIT,COMPO,ANDD	;?compile only lexicon bits
		DW	ABORQ
		DB	13,' compile only'
		DW	EXECU,EXIT		;execute defined word
INTE1:		DW	TNUMB,ATEXE		;convert a number
		DW	QBRAN,INTE2
		DW	EXIT
INTE2:		DW	THROW			;error

;   [		( -- )
;		Start the text interpreter.

		$COLON	IMEDD+1,'[',LBRAC
		DW	DOLIT,INTER,TEVAL,STORE,EXIT

;   .OK 	( -- )
;		Display 'ok' only while interpreting.

		$COLON	3,'.OK',DOTOK
		DW	DOLIT,INTER,TEVAL,AT,EQUAL
		DW	QBRAN,DOTO1
		DW	DOTQP
		DB	3,' ok'
DOTO1:		DW	CR,S_O,EXIT

;   ?STACK	( -- )
;		Abort if the data stack underflows.

		$COLON	6,'?STACK',QSTAC
		DW	DEPTH,ZLESS		;check only for underflow
		DW	ABORQ
		DB	10,' underflow',0
		DW	EXIT

;   EVAL	( -- )
;		Interpret the input stream.

		$COLON	4,'EVAL',EVAL
EVAL1:		DW	TOKEN,DUPP,CAT		;?input stream empty
		DW	QBRAN,EVAL2
		DW	TEVAL,ATEXE,QSTAC	;evaluate input, check stack
		DW	BRAN,EVAL1
EVAL2:		DW	DROP,TPROM,ATEXE,EXIT	;prompt

;; Shell

;   PRESET	( -- )
;		Reset data stack pointer and the terminal input buffer.

		$COLON	6,'PRESET',PRESE
		DW	SZERO,AT,SPSTO
		DW	DOLIT,TIBB,NTIB,CELLP,STORE,EXIT

;   xio 	( addr. a addr. -- )
;		Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

		$COLON	COMPO+3,'xio',XIO
		DW	DOLIT,ACCEP,TEXPE,DSTOR
		DW	TECHO,DSTOR,EXIT

;   HAND	( -- )
;		Select I/O vectors for terminal interface.

		$COLON	4,'HAND',HAND
		DW	DOLIT,DOTOK,DOLIT,EMIT
		DW	DOLIT,KTAP,XIO,EXIT

;  fTAP 	( bot eot cur c -- bot eot cur )
;		Process a key stroke ,CR or CTRL_Z.

		$COLON	4,'fTAP',FTAP
		DW	DUPP,DOLIT,CRR,XORR
		DW	QBRAN,FTAP2
		DW	DOLIT,CTRL_Z,XORR
		DW	QBRAN,FTAP1
		DW	BLANK,TAP,EXIT
FTAP1:		DW	HAND,DOTOK,EXIT
FTAP2:		DW	DROP,SWAP,DROP,DUPP,EXIT

;   FILE	( -- )
;		Select I/O vectors for file download.

		$COLON	4,'FILE',FILE
		DW	DOLIT,PACE,DOLIT,DROP
		DW	DOLIT,FTAP,XIO,EXIT

;   I/O 	( -- addr.)
;		Array to store default I/O vectors.

		$VAR	3,'I/O',ISLO
		DB	0FDH,0FDH		; NOP NOP
		DW	QRX,TXSTO		;default I/O vectors

;   CONSOLE	( -- )
;		Initiate terminal interface.

		$COLON	7,'CONSOLE',CONSO
		DW	ISLO,DAT,TQKEY,DSTOR	;restore default I/O device
		DW	HAND,EXIT		;keyboard input

;   QUIT	( -- )
;		Reset return stack pointer and start text interpreter.

		$COLON	4,'QUIT',QUIT
		DW	RZERO,AT,RPSTO		;reset return stack pointer
QUIT1:		DW	LBRAC			;start interpretation
QUIT2:		DW	QUERY			;get input
		DW	DOLIT,EVAL,CATCH,QDUP	;evaluate input
		DW	QBRAN,QUIT2		;continue till error
		DW	TPROM,AT,TOR		;save input device
		DW	CONSO,NULLS,OVER,XORR	;?display error message
		DW	QBRAN,QUIT3
		DW	SPACE,COUNT,TYPES	;error message
		DW	DOTQP
		DB	5,'  ? ',CTRL_Z         ;error prompt
QUIT3:		DW	RFROM,DOLIT,DOTOK,XORR	;?file input
		DW	QBRAN,QUIT4
		DW	DOLIT,ERR,EMIT		;file error, tell host
QUIT4:		DW	PRESE			;some cleanup
		DW	BRAN,QUIT1

;; The compiler

;   '           ( -- ca )
;		Search context vocabularies for the next word in input stream.

		$COLON	1,"'",TICK
		DW	TOKEN,NAMEQ		;?defined
		DW	QBRAN,TICK1
		DW	EXIT			;yes, push code address
TICK1:		DW	THROW			;no, error

;   ALLOT	( n -- )
;		Allocate n bytes to the code dictionary.

		$COLON	5,'ALLOT',ALLOT
		DW	CP,PSTOR,EXIT		;adjust code pointer

;   RALLOT	( n -- )
;		Allocate n bytes to RAM.

		$CODE	6,'RALLOT',RALLOT
		DB	0C8H,22H		;POP BX
		DB	64H,22H,28H		;ADD VP,BX
						;adjust RAM pointer
		$NEXT

;   ,		( w -- )
;		Compile an integer into the code dictionary.

		$COLON	1,',',COMMA
		DW	HERE,DUPP,CELLP 	;cell boundary
		DW	CP,STORE,STORE,EXIT	;adjust code pointer and compile

;   C,		( c -- )
;		Compile a character into the code dictionary.

		$COLON	2,'C,',CCOMMA
		DW	HERE,DUPP,INCR
		DW	CP,STORE,CSTOR,EXIT

;   R,		( w -- )
;		Compile an integer into RAM.

		$CODE	2,'R,',RCOMM
		DB	0CCH,22H		;POP BX
		DB	0C2H,29H,22H		;ST BX,[VP]+
		$NEXT

;   RC, 	( c -- )
;		Compile a character into RAM.

		$CODE	3,'RC,',RCCOMM
		DB	0CCH,22H		;POP BX
		DB	0C6H,29H,22H		;STB BL,[VP]+
		$NEXT

;   [COMPILE]	( -- ; <string> )
;		Compile the next immediate word into code dictionary.

		$COLON	IMEDD+9,'[COMPILE]',BCOMP
		DW	TICK,COMMA,EXIT

;   COMPILE	( -- )
;		Compile the next address in colon list to code dictionary.

		$COLON	COMPO+7,'COMPILE',COMPI
		DW	RFROM,DUPP,AT,COMMA	;compile address
		DW	CELLP,TOR,EXIT		;adjust return address

;   LITERAL	( w -- )
;		Compile tos to code dictionary as an integer literal.

		$COLON	IMEDD+7,'LITERAL',LITER
		DW	COMPI,DOLIT,COMMA,EXIT

;   $,"         ( -- )
;		Compile a literal string up to next " .

		$COLON	3,'$,"',STRCQ
		DW	DOLIT,'"',WORDD         ;move string to code dictionary
		DW	COUNT,PLUS,ALGND	;calculate aligned end of string
		DW	CP,STORE,EXIT		;adjust the code pointer

;   RECURSE	( -- )
;		Make the current word available for compilation.

		$COLON	IMEDD+7,'RECURSE',RECUR
		DW	LAST,AT,NAMET,COMMA,EXIT

;; Structures

;   FOR 	( -- addr.)
;		Start a FOR-NEXT loop structure in a colon definition.

		$COLON	IMEDD+3,'FOR',FOR
		DW	COMPI,TOR,HERE,EXIT

;   BEGIN	( -- addr.)
;		Start an infinite or indefinite loop structure.

		$COLON	IMEDD+5,'BEGIN',BEGIN
		DW	HERE,EXIT

;   NEXT	( addr. -- )
;		Terminate a FOR-NEXT loop structure.

		$COLON	IMEDD+4,'NEXT',NEXT
		DW	COMPI,DONXT,COMMA,EXIT

;   DO		( -- loop address )
;		Start a DO ... LOOP ( or DO ... +LOOP) structure.

		$COLON	IMEDD+2,'DO',DO
		DW	COMPI,PARDO,HERE,EXIT

;   LOOP	( loop address -- )
;		Terminate a DO...LOOP Structure.

		$COLON IMEDD+4,'LOOP',LOOP
		DW	COMPI,PLOP,COMMA,EXIT

;  +LOOP	( loop address -- )
;		Terminate a DO...+LOOP Structure.

		$COLON IMEDD+5,'+LOOP',PLOOP
		DW	COMPI,PPLOP,COMMA,EXIT

;   LEAVE	( -- )

		$COLON IMEDD+5,'LEAVE',LEAVE
		DW	COMPI,PLEAVE,EXIT

;   UNTIL	( addr. -- )
;		Terminate a BEGIN-UNTIL indefinite loop structure.

		$COLON	IMEDD+5,'UNTIL',UNTIL
		DW	COMPI,QBRAN,COMMA,EXIT

;   AGAIN	( addr. -- )
;		Terminate a BEGIN-AGAIN infinite loop structure.

		$COLON	IMEDD+5,'AGAIN',AGAIN
		DW	COMPI,BRAN,COMMA,EXIT

;   IF		( -- A )
;		Begin a conditional branch structure.

		$COLON	IMEDD+2,'IF',IFF
		DW	COMPI,QBRAN,HERE
		DW	DOLIT,0,COMMA,EXIT

;   AHEAD	( -- A )
;		Compile a forward branch instruction.

		$COLON	IMEDD+5,'AHEAD',AHEAD
		DW	COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;   REPEAT	( A addr. -- )
;		Terminate a BEGIN-WHILE-REPEAT indefinite loop.

		$COLON	IMEDD+6,'REPEAT',REPEA
		DW	AGAIN,HERE,SWAP,STORE,EXIT

;   THEN	( A -- )
;		Terminate a conditional branch structure.

		$COLON	IMEDD+4,'THEN',THENN
		DW	HERE,SWAP,STORE,EXIT

;   AFT 	( addr. -- addr. A )
;		Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

		$COLON	IMEDD+3,'AFT',AFT
		DW	DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE	( A -- A )
;		Start the false clause in an IF-ELSE-THEN structure.

		$COLON	IMEDD+4,'ELSE',ELSEE
		DW	AHEAD,SWAP,THENN,EXIT

;   WHILE	( addr. -- A addr.)
;		Conditional branch out of a BEGIN-WHILE-REPEAT loop.

		$COLON	IMEDD+5,'WHILE',WHILE
		DW	IFF,SWAP,EXIT

;   CASE	( -- addr. )
;		Begin CASE-OF-ENDOF-ENDCASE structure.

		$COLON	IMEDD+4,'CASE',CASE
		DW	CSP,AT,STCSP,EXIT

;   OF		( n1,n2 -- n1(dismatch)/(match) )
;		Compile CASE-OF-ENDOF-ENDCASE structure.

		$COLON	IMEDD+2,'OF',OF
		DW	COMPI,OVER,COMPI,EQUAL
		DW	COMPI,QBRAN,HERE,DOLIT,0,COMMA
		DW	COMPI,DROP,EXIT

;   ENDOF	( -- <runtime> // addr1 n1 -- addr2 n2 <compiling>)
;		Compile CASE-OF-ENDOF-ENDCASE structure.

		$COLON	IMEDD+5,'ENDOF',ENDOF
		DW	COMPI,BRAN,HERE,DOLIT,0,COMMA
		DW	SWAP,THENN,EXIT

;   ENDCASE	( -- )
;		Terminate CASE-OF-ENDOF-ENDCASE structure.

		$COLON	IMEDD+7,'ENDCASE',ECAS
		DW	COMPI,DROP
ECAS1:		DW	SPAT,CSP,AT,EQUAL,EQUAL_ZERO
		DW	QBRAN,ECAS2
		DW	THENN
		DW	BRAN,ECAS1
ECAS2:		DW	CSP,STORE,EXIT

;   ABORT"      ( -- ; <string> )
;		Conditional abort with an error message.

		$COLON	IMEDD+6,'ABORT"',ABRTQ
		DW	COMPI,ABORQ,STRCQ,EXIT

;   $"          ( -- ; <string> )
;		Compile an inline string literal.

		$COLON	IMEDD+2,'$"',STRQ
		DW	COMPI,STRQP,STRCQ,EXIT

;   ."          ( -- ; <string> )
;		Compile an inline string literal to be typed out at run time.

		$COLON	IMEDD+2,'."',DOTQ
		DW	COMPI,DOTQP,STRCQ,EXIT

;; Name compiler

;   ?UNIQUE	( addr. -- addr.)
;		Display a warning message if the word already exists.

		$COLON	7,'?UNIQUE',UNIQU
		DW	DUPP,NAMEQ		;?name exists
		DW	QBRAN,UNIQ1
		DW	DOTQP			;redefinitions are OK
		DB	7,' reDef '             ;but the user should be warned
		DW	OVER,COUNT,TYPES	;just in case its not planned
UNIQ1:		DW	DROP,EXIT

;   $,n 	( na -- )
;		Build a new dictionary name using the string at na.

		$COLON	3,'$,n',SNAME
		DW	DUPP,CAT		;?null input
		DW	QBRAN,PNAM1
		DW	UNIQU			;?redefinition
		DW	DUPP,LAST,STORE 	;save na for vocabulary link
		DW	HERE,ALGND,SWAP 	;align code address
		DW	CELLM			;link address
		DW	CRRNT,AT,AT,OVER,STORE
		DW	CELLM,DUPP,NP,STORE	;adjust name pointer
		DW	STORE,EXIT		;save code pointer
PNAM1:		DW	STRQP
		DB	5,' name'               ;null input
		DW	THROW

;; FORTH compiler

;   $COMPILE	( addr. -- )
;		Compile next word to code dictionary as a token or literal.

		$COLON	8,'$COMPILE',SCOMP
		DW	NAMEQ,QDUP		;?defined
		DW	QBRAN,SCOM2
		DW	AT,DOLIT,IMEDD,ANDD	;?immediate
		DW	QBRAN,SCOM1
		DW	EXECU,EXIT		;its immediate, execute
SCOM1:		DW	COMMA,EXIT		;its not immediate, compile
SCOM2:		DW	TNUMB,ATEXE		;try to convert to number
		DW	QBRAN,SCOM3
		DW	LITER,EXIT		;compile number as integer
SCOM3:		DW	THROW			;error

;   OVERT	( -- )
;		Link a new word into the current vocabulary.

		$COLON	5,'OVERT',OVERT
		DW	LAST,AT,CRRNT,AT,STORE,EXIT

;   ;		( -- )
;		Terminate a colon definition.

		$COLON	IMEDD+COMPO+1,';',SEMIS
		DW	COMPI,EXIT,LBRAC,OVERT,EXIT

;   ]		( -- )
;		Start compiling the words in the input stream.

		$COLON	1,']',RBRAC
		DW	DOLIT,SCOMP,TEVAL,STORE,EXIT

;   call,	( ca -- )
;		Assemble a call instruction to ca.

		$COLON	5,'call,',CALLC
		DW	DOLIT,LCALLL,COMMA,HERE  ;Direct Threaded Code
		DW	CELLP,SUBB,COMMA,EXIT	 ;DTC 8098 relative call

;   :		( -- ; <string> )
;		Start a new colon definition using next word as its name.

		$COLON	1,':',COLON
		DW	TOKEN,SNAME,DOLIT,DOLST
		DW	CALLC,RBRAC,EXIT

;   IMMEDIATE	( -- )
;		Make the last compiled word an immediate word.

		$COLON	9,'IMMEDIATE',IMMED
		DW	DOLIT,IMEDD,LAST,AT,AT,ORR
		DW	LAST,AT,STORE,EXIT

;; Defining words

;   CONSTANT	( u -- ; <string> )
;		Compile a new constant.

		$COLON	8,'CONSTANT',CONST
		DW	TOKEN,SNAME,OVERT
		DW	DOLIT,DOCONST,CALLC
		DW	COMMA,EXIT

;   USER	( u -- ; <string> )
;		Compile a new user variable.

		$COLON	4,'USER',USER
		DW	TOKEN,SNAME,OVERT
		DW	DOLIT,DOUSE,CALLC
		DW	COMMA,EXIT

;   CREATE     ( -- ; <string> )
;		Compile a new array entry without allocating code space.

		$COLON	6,'CREATE',CREAT
		DW	TOKEN,SNAME,OVERT
		DW	DOLIT,DOVAR,CALLC
		DW	DOLIT,0,COMMA		;compitable with others.
		DW	EXIT

;   RCREATE	( -- ; <string> )
;		Compile a new array entry without allocating RAM space.

		$COLON	7,'RCREATE',RCREAT
		DW	TOKEN,SNAME,OVERT
		DW	DOLIT,RDOVAR,CALLC
		DW	RHERE,COMMA		;compile available RAM addr.
		DW	EXIT

;   VARIABLE	( -- ; <string> )
;		Compile a new variable in RAM initialized to 0.

		$COLON	8,'VARIABLE',VARIA
		DW	RCREAT,DOLIT,0,RCOMM,EXIT

;   DOES>	( -- ; <string> )
;		Beginning of high level definitions.

		$COLON	COMPO+5,'DOES>',DOEST
		DW	RFROM,LAST,AT,NAMET	;tos=addr.,CFA
		DW	CELLP,DUPP,DUPP,CELLP	;tos=addr.,CFA+2,CFA+2,CFA+4
		DW	DOLIT,DODOES		;
		DW	SWAP,SUBB,SWAP,STORE	;tos=addr.,CFA+2
		DW	CELLP,STORE
		DW	EXIT

;   SBR         ( -- ; <string> )
;               Beginning of subroutine.

                $COLON  IMEDD+3,'SBR',SBR
                DW      TOKEN,SNAME,OVERT
                DW      DOLIT,DOVAR,CALLC
                DW      DOLIT,0,COMMA
                DW      DOLIT,0F2FDH,COMMA      ;NOP PUSHF
                DW      DOLIT,1CC8H,COMMA       ;PUSH RP
                DW      DOLIT,1AC8H,COMMA       ;PUSH IP
                DW      DOLIT,20C8H,COMMA       ;PUSH AX
                DW      DOLIT,22C8H,COMMA       ;PUSH BX
                DW      DOLIT,24C8H,COMMA       ;PUSH CX
                DW      DOLIT,26C8H,COMMA       ;PUSH DX
                DW      DOLIT,DOLST,CALLC
                DW      RBRAC,EXIT

;   END-SBR     ( -- )
;               Terminal the compile.

                $COLON  COMPO+IMEDD+7,'END-SBR',ENDSBR
                DW      LBRAC,EXIT

;   ;CODE	( -- ; <string> )
;		Beginning of assmbly level definitions.

                $COLON  COMPO+IMEDD+5,';CODE',SCODE
		DW	QCSP
		DW	DOLIT,PSCODE,COMMA
		DW	LBRAC,EXIT

;   CODE        ( --; <string> )
;               Beginning of CODE- def.

                $COLON  IMEDD+4,'CODE',CODD
                DW      STCSP
                DW      TOKEN,SNAME,OVERT
                DW      ASSMB,FRCLR,RESETF
                DW      EXIT

;; Vocabulary words

;   VOCABULARY	( -- )
;		Define a new vocabulary.

		$COLON	10,'VOCABULARY',VOCA
		DW	CREAT
		DW	RHERE,COMMA
		DW	DOLIT,0,RCOMM
		DW	RHERE
		DW	VOCLK,AT,RCOMM
		DW	LAST,AT,NAMET,RCOMM
		DW	VOCLK,STORE
		DW	DOEST
VOCA1:		DW	DOVOC
		DW	EXIT

;   DEFINITIONS ( -- )
;		Link subsequent definitions to the context vocabulary.

		$COLON	11,'DEFINITIONS',DEFN
		DW	CNTXT,AT
		DW	CRRNT,STORE
		DW	EXIT

;   ONLY	( -- )
;		Begin a only-also structure.

		$COLON	4,'ONLY',ONLY
                DW      CNTXT,DOLIT,8,ERASE
		DW	DOLIT,VPP0,CNTXT,STORE
		DW	DOLIT,VPP0,CNTXT,CELLP,STORE
		DW	EXIT

;   ALSO	( --)
;		Treminate a ONLY-ALSO structure.

		$COLON	4,'ALSO',ALSO
		DW	CNTXT,DUPP,CELLP
		DW	DOLIT,16,CMVET
		DW	EXIT
;; Tools

;   _TYPE	( b u -- )
;		Display a string. Filter non-printing characters.

		$COLON	5,'_TYPE',UTYPE
		DW	TOR			;start count down loop
		DW	BRAN,UTYP2		;skip first pass
UTYP1:		DW	DUPP,CAT,TCHAR,EMIT	;display only printable
		DW	INCR			;increment address
UTYP2:		DW	DONXT,UTYP1		;loop till done
		DW	DROP,EXIT

;   dm+ 	( addr. u -- addr.)
;		Dump u bytes from , leaving a+u on the stack.

		$COLON	3,'dm+',DUMPP
		DW	OVER,DOLIT,4,UDOTR	;display address
		DW	SPACE,TOR		;start count down loop
		DW	BRAN,PDUM2		;skip first pass
PDUM1:		DW	DUPP,CAT,DOLIT,3,UDOTR	;display numeric data
		DW	INCR			;increment address
PDUM2:		DW	DONXT,PDUM1		;loop till done
		DW	EXIT

;   DUMP	( addr. u -- )
;		Dump u bytes from a, in a formatted manner.

		$COLON	4,'DUMP',DUMP
		DW	BASE,AT,TOR,HEX 	;save radix, set hex
		DW	DOLIT,16,SLASH		;change count to lines
		DW	TOR			;start count down loop
DUMP1:		DW	CR,DOLIT,16,DDUP,DUMPP	;display numeric
		DW	ROT,ROT
		DW	DOLIT,2,SPACS,UTYPE	;display printable characters
		DW	DONXT,DUMP1		;loop till done
		DW	DROP,RFROM,BASE,STORE	;restore radix
		DW	EXIT

;   .S		( ... -- ... )
;		Display the contents of the data stack.

		$COLON	2,'.S',DOTS
		DW	CR,DEPTH		;stack depth
		DW	TOR			;start count down loop
		DW	BRAN,DOTS2		;skip first pass
DOTS1:		DW	RAT,PICK,DOT		;index stack, display contents
DOTS2:		DW	DONXT,DOTS1		;loop till done
		DW	DOTQP
		DB	4,' <sp',0
		DW	EXIT

;   !CSP	( -- )
;		Save stack pointer in CSP for error checking.

		$COLON	4,'!CSP',STCSP
		DW	SPAT,CSP,STORE,EXIT	;save pointer

;   ?CSP	( -- )
;		Abort if stack pointer differs from that saved in CSP.

		$COLON	4,'?CSP',QCSP
		DW	SPAT,CSP,AT,XORR	;compare pointers
		DW	ABORQ			;abort if different
		DB	6,'stacks',0
		DW	EXIT

;   >NAME	( ca -- na | F )
;		Convert code address to a name address.

		$COLON	5,'>NAME',TNAME
		DW	CRRNT			;vocabulary link
TNAM1:		DW	CELLP,AT,QDUP		;check all vocabularies
		DW	QBRAN,TNAM4
		DW	DDUP
TNAM2:		DW	AT,DUPP 		;?last word in a vocabulary
		DW	QBRAN,TNAM3
		DW	DDUP,NAMET,XORR 	;compare
		DW	QBRAN,TNAM3
		DW	CELLM			;continue with next word
		DW	BRAN,TNAM2
TNAM3:		DW	SWAP,DROP,QDUP
		DW	QBRAN,TNAM1
		DW	SWAP,DROP,SWAP,DROP,EXIT
TNAM4:		DW	DROP,DOLIT,0,EXIT

;   .ID 	( na -- )
;		Display the name at address.

		$COLON	3,'.ID',DOTID
		DW	QDUP			;if zero no name
		DW	QBRAN,DOTI1
		DW	COUNT,DOLIT,01FH,ANDD	;mask lexicon bits
		DW	UTYPE,EXIT		;display name string
DOTI1:		DW	DOTQP
		DB	9,' {noName}'
		DW	EXIT

;   SEE 	( -- ;<string> )
;		A simple decompiler.

		$COLON	3,'SEE',SEE
		DW	TICK			;starting address
		DW	CR,CELLP
SEE1:		DW	CELLP,DUPP,DUPP,AT,DUPP ;?does it contian a zero
		DW	QBRAN,SEE2
		DW	TNAME			;?is it a name
SEE2:		DW	QDUP			;name address or zero
		DW	QBRAN,SEE3
		DW	SPACE,DOTID		;display name
		DW	BRAN,SEE4
SEE3:		DW	DUPP,AT,UDOT		;display number
SEE4:		DW	AT,DOLIT,EXIT,EQUAL	;display controller
		DW	QBRAN,SEE1
		DW	DROP,EXIT

;   WORDS	( -- )
;		Display the names in the context vocabulary.

		$COLON	5,'WORDS',WORDS
		DW	CR,CNTXT,AT		;only in context
WORS1:		DW	AT,QDUP 		;?at end of list
		DW	QBRAN,WORS2
		DW	DUPP,SPACE,DOTID	;display a name
		DW	CELLM			;user control
		DW	BRAN,WORS1
WORS2:		DW	EXIT

;   VOCS	( -- )
;		Display all vocabularies.

		$COLON	4,'VOCS',VOCS
		DW	CR,DOTQP
		DB	23,' Defined vocabularies: '
		DW	VOCLK,AT
VOCS1:		DW	DUPP
		DW	CELLP,AT,TNAME
		DW	SPACE,DOTID
		DW	AT,DUPP
		DW	QBRAN,VOCS2
		DW	BRAN,VOCS1
VOCS2:		DW	DROP,EXIT

;   ORDER	( -- )
;		Display vocabulary search order.

		$COLON	5,'ORDER',ORDER
		DW	CR,DOTQP
		DB	9,'Context: '
		DW	CNTXT
ORDER1: 	DW	DUPP,AT,DUPP
		DW	QBRAN,ORDER2
		DW	CELLP,CELLP
		DW	AT,TNAME
		DW	SPACE,DOTID
		DW	CELLP
		DW	BRAN,ORDER1
ORDER2: 	DW	DROP,DROP
		DW	CR,DOTQP
		DB	9,'Current: '
		DW	CRRNT,AT
		DW	CELLP,CELLP
		DW	AT,TNAME
		DW	SPACE,DOTID
		DW	EXIT

;   LOAD	( -- )
;		File handle.

		$COLON	4,'LOAD',FLOAD
		DW	DOLIT,FSRM,EMIT
		DW	BLANK,PARSE,TYPES
		DW	DOLIT,CTRL_Z,EMIT
		DW	FILE,EXIT

;   MAKE        ( -- ;<string> )
;               Make next words as boot routine.

                $COLON  4,'MAKE',MAKE
                DW      TICK
                DW      DOLIT,0BDH,DOLIT,2018H,CSTOR
                DW      DOLIT,0A1H,DOLIT,2080H,CSTOR
                DW      SZERO,AT
                DW      DOLIT,2081H,STORE
                DW      DOLIT,18H,DOLIT,2083H,CSTOR
                DW      DOLIT,0A1H,DOLIT,2084H,CSTOR
                DW      RZERO,AT
                DW      DOLIT,2085H,STORE
                DW      DOLIT,1CH,DOLIT,2087H,CSTOR
                DW      DOLIT,0A1H,DOLIT,2088H,CSTOR
                DW      UPZERO,AT
                DW      DOLIT,2089H,STORE
                DW      DOLIT,1EH,DOLIT,208BH,CSTOR
                DW      DOLIT,0A1H,DOLIT,208CH,CSTOR
                DW      DOLIT,208DH,STORE
                DW      DOLIT,1AH,DOLIT,208FH,CSTOR
                DW      EXIT

;   epwr        ( addr1 addr2 addr3 type --- f )
;               EPROM writer. If success , return True, else False.

                $CODE   4,'epwr',EPWRR
                DB      0CCH,32H                ;POP TX
                DB      0CCH,30H                ;POP EX
                DB      0CCH,26H                ;POP DX
                DB      0CCH,22H                ;POP BX
                DB      0C8H,22H                ;PUSH BX
                DB      0C8H,30H                ;PUSH EX
EPWRR1:         DB      0B1H,80H,20H            ;LDB AL,#80H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,0CEH,20H           ;LDB AL,#0CEH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C7H,01H,0FCH,0FFH,30H  ;STB EL,0FFFCH[0]
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
                DB      0B1H,4EH,20H            ;LDB AL,#4EH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0B2H,23H,21H            ;LDB AH,[BX]+
                DB      99H,0FFH,21H            ;CMPB AH,#0FFH
                DB      0D7H,03H             ;JNE  EPWRR11
                DB      0E7H                    ;LJMP EPWRR5
                DW      EPWRR5-$-2
EPWRR11:        DB      11H,24H                 ;CLRB CL
                DB      0C7H,01H,0FCH,0FFH,21H  ;STB AH,0FFFCH[0]
                DB      0B1H,4EH,20H            ;LDB AL,#4EH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0FDH                    ;NOP
                DB      0FDH                    ;NOP
EPWRR2:         DB      0B1H,4AH,20H            ;LDB AL,#4AH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C8H,22H                ;PUSH BX
                DB      0B1H,01H,23H            ;LDB BH,#01H
                DB      0EFH                    ;LCALL  DELAY
                DW      DELAY-$-2
                DB      0CCH,22H                ;POP  BX
                DB      75H,01H,24H             ;ADDB CL,#01H
                DB      0B1H,4EH,20H            ;LDB AL,#4EH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      99H,0FH,24H             ;CMPB CL,#0FH
                DB      0D7H,03H             ;JNE EPWRR21
                DB      0E7H                    ;LJMP EPWRRA
                DW      EPWRRA-$-2
EPWRR21:        DB      0B1H,90H,20H            ;LDB AL,#90H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,46H,20H            ;LDB AL,#46H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
                DB      99H,03H,32H             ;CMPB TL,#03H
                DB      0DFH,08H             ;JE  EPWRR3
                DB      95H,40H,31H             ;XOR EH,#40H
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
EPWRR3:         DB      0B1H,42H,20H            ;LDB AL,#42H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0FDH                    ;NOP
                DB      0B3H,01H,0FCH,0FFH,25H  ;LDB CH,0FFFCH[0]
                DB      0B1H,4EH,20H            ;LDB AL,#4EH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      99H,03H,32H             ;CMPB TL,#03H
                DB      0DFH,08H             ;JE  EPWRR30
                DB      95H,40H,31H             ;XOR EH,#40H
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
EPWRR30:        DB      0B1H,80H,20H            ;LDB AL,#80H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,4EH,20H            ;LDB AL,#4EH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
                DB      0FDH                    ;NOP
                DB      0C7H,01H,0FCH,0FFH,21H  ;STB AH,0FFFCH[0]
                DB      98H,25H,21H             ;CMPB AH,CH
                DB      0DFH,03H              ;JE EPWRR31
                DB      0E7H                    ;LJMP EPWRR2
                DW      EPWRR2-$-2
EPWRR31:        DB      0B1H,4AH,20H            ;LDB AL,#4AH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
EPWRR4:         DB      0C8H,22H                ;PUSH BX
                DB      0B1H,01H,23H            ;LDB BH,#01H
                DB      0EFH                    ;LCALL DELAY
                DW      DELAY-$-2
                DB      0CCH,22H                ;POP  BX
                DB      0E0H,24H,0F3H         ;DJNZ CL,EPWRR4
EPWRR5:         DB      07H,30H                 ;INC EX
                DB      88H,26H,22H             ;CMP BX,DX
                DB      0D9H,03H              ;JH EPWRR6
                DB      0E7H                    ;LJMP EPWRR1
                DW      EPWRR1-$-2
EPWRR6:         DB      0CCH,26H                ;POP DX
                DB      0CCH,22H                ;POP BX
EPWRR7:         DB      0B1H,80H,20H            ;LDB AL,#80H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,0C6H,20H           ;LDB AL,#0C6H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C7H,01H,0FCH,0FFH,26H  ;STB DL,0FFFCH[0]
                DB      0C7H,01H,0FDH,0FFH,27H  ;STB DH,0FFFDH[0]
                DB      0FDH                    ;NOP
                DB      0B1H,46H,20H            ;LDB AL,#46H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0B1H,90H,20H            ;LDB AL,#90H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,46H,20H            ;LDB AL,#46H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C7H,01H,0FDH,0FFH,27H  ;STB DH,0FFFDH[0]
                DB      99H,03H,32H             ;CMPB TL,#03H
                DB      0DFH,08H              ;JE  EPWRR8
                DB      95H,40H,27H             ;XOR EH,#40H
                DB      0C7H,01H,0FDH,0FFH,27H  ;STB EH,0FFFDH[0]
EPWRR8:         DB      0B1H,42H,20H            ;LDB AL,#42H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0FDH                    ;NOP
                DB      0B3H,01H,0FCH,0FFH,25H  ;LDB CH,0FFFCH[0]
                DB      0B2H,23H,21H            ;LDB AH,[BX]+
                DB      98H,25H,21H             ;CMPB AH,CH
                DB      0D7H,0DH                ;JNE EPWRR9
                DB      07H,26H                 ;INC DX
                DB      88H,30H,26H             ;CMP DX,EX
                DB      0D1H,9EH                ;JNH EPWRR7
                DB      0C9H,0FFH,0FFH          ;PUSH #0FFFFH
                DB      0E7H                    ;LJMP EPWRRC
                DW      EPWRRC-$-2
EPWRR9:         DB      0C9H,00H,00H            ;PUSH #0
                DB      0E7H                    ;LJMP EPWRRC
                DW      EPWRRC-$-2
EPWRRA:         DB      0B1H,4AH,20H            ;LDB AL,#4AH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C8H,22H                ;PUSH BX
                DB      0B1H,32H,23H            ;LDB BH,#32H
                DB      0EFH                    ;LCALL DELAY
                DW      DELAY-$-2
                DB      0CCH,22H                ;POP BX
                DB      0B1H,4EH,20H            ;LDB AL,#4EH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0B1H,90H,20H            ;LDB AL,#90H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,46H,20H            ;LDB AL,#46H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
                DB      99H,03H,32H             ;CMPB TL,#03H
                DB      0DFH,08H              ;JE  EPWRRB
                DB      95H,40H,31H             ;XOR EH,#40H
                DB      0C7H,01H,0FDH,0FFH,31H  ;STB EH,0FFFDH[0]
EPWRRB:         DB      0B1H,42H,20H            ;LDB AL,#42H
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                DB      0FDH                    ;NOP
                DB      0B3H,01H,0FCH,0FFH,25H  ;LDB CH,0FFFCH[0]
                DB      98H,25H,21H             ;CMPB AH,CH
                DB      0D7H,03H              ;JNE  EPWRRB1
                DB      0E7H                    ;LJMP EPWRR5
                DW      EPWRR5-$-2
EPWRRB1:        DB      65H,04H,00H,18H         ;ADD SP,#04H
                DB      0C9H,00H,00H            ;PUSH #0
EPWRRC:         DB      0B1H,80H,20H            ;LDB AL,#80H
                DB      0C7H,01H,0FFH,0FFH,20H  ;STB AL,0FFFFH[0]
                DB      0B1H,0FH,20H            ;LDB AL,#0FH
                DB      0C7H,01H,0FEH,0FFH,20H  ;STB AL,0FFFEH[0]
                $NEXT

;; Hardware reset

;   VER 	( -- n )
;		Return the version number of this implementation.

		$COLON	3,'VER',VERSN
		DW	DOLIT,VER*256+EXT,EXIT

;   hi		( -- )
;		Display the sign-on message of eForth.

		$COLON	2,'hi',HI
		DW	STOIO,CR,DOTQP		;initialize I/O
		DB	13,'Forth-98 v'         ;model
		DB	VER+'0','.',EXT+'0'     ;version
		DW	CR,S_O,EXIT

;   'BOOT       ( -- addr.)
;		The application startup vector.

		$VAR	5,"'BOOT",TBOOT
		DB	0FDH,0FDH		; NOP NOP
		DW	HI			;application to boot

;   COLD	( -- )
;		The hilevel cold start sequence.

		$COLON	4,'COLD',COLD
COLD1:		DW	DOLIT,UZERO,DOLIT,UPP
                DW      DOLIT,ULAST-UZERO,CMOVE ;initialize user area
		DW	PRESE			;initialize data stack and TIB
		DW	TBOOT,ATEXE		;application boot
		DW	DOLIT,LASTN		;the follow is initialization .
		DW	DOLIT,VPP0,STORE
		DW	DOLIT,0
		DW	DOLIT,VPP0+2,STORE
		DW	DOLIT,FORTH
		DW	DOLIT,VPP0+4,STORE
                DW      DOLIT,ALASTN
		DW	DOLIT,VPP0+6,STORE
		DW	DOLIT,VPP0+2
		DW	DOLIT,VPP0+8,STORE
		DW	DOLIT,ASSMB
		DW	DOLIT,VPP0+10,STORE
		DW	DOLIT,VPP0+8,VOCLK,STORE
		DW	FORTH,CNTXT,AT,DUPP	;initialize search order
		DW	CNTXT,CELLP,STORE,DUPP	;  : forth,forth,0
		DW	CRRNT,DSTOR
		DW	QUIT			;start interpretation
		DW	BRAN,COLD1		;just in case

;===============================================================

LASTN		EQU	_NAME+4 		;last name address in name dictionary

;; ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

;; asm96.seq

ANAMEE          EQU     _NAME
ACODEE          EQU     0C000H

;  Redefine assembly variables

_LINK = 0                       ; force a null link
_NAME = ANAMEE                  ; redefine name pointer
_CODE = ACODEE                  ; redefine code pointer
                
ORG     ACODEE

;; used for jump instructions

;   FRTABLE     ( -- )
;               Forward refrence table

                $USER   7,'FRTABLE',FRTABLE
                _USER = _USER+39*CELLL

;   LTABLE      ( -- )
;               Local label table

                $USER   6,'LTABLE',LTABLE
                _USER = _USER+19*CELLL

;   FRCLR       ( -- )
;               Initialize forward refrence table

                $COLON  5,'FRCLR',FRCLR
                DW      FRTABLE,DOLIT,28H,ERASE
                DW      LTABLE,DOLIT,14H,ERASE
                DW      EXIT

;   FRRES       ( -- )
;               Resolve forward refrences

                $COLON  5,'FRRES',FRRES
                DW      DOLIT,0AH,DOLIT,0
                DW      PARDO
FRRES1:         DW      I,DOLIT,4,STAR
                DW      FRTABLE,PLUS,DUPP,AT
                DW      DOLIT,2,PICK,EQUAL
                DW      QBRAN,FRRES8
                DW      DUPP,CELLP,AT,DUPP
                DW      CAT,DOLIT,0E7H
                DW      ANDD,DOLIT,0E7H,EQUAL
                DW      QBRAN,FRRES2
                DW      DUPP,DOLIT,3,PLUS
                DW      HERE,SWAP,SUBB
                DW      SWAP,INCR,STORE         ; 16 bit
                DW      BRAN,FRRES7
FRRES2:         DW      DUPP,CAT,DUPP
                DW      DOLIT,27H,EQUAL,SWAP
                DW      DOLIT,2FH,EQUAL,ORR
                DW      QBRAN,FRRES4
                DW      DUPP,DUPP,DUPP,CELLP
                DW      HERE,SWAP,SUBB,DUPP
                DW      ABSS,DOLIT,07FFH,GREAT
                DW      QBRAN,FRRES3
                DW      ABORQ
                DB      13,' Out of range'
                DW      BRAN,FRRES4
FRRES3:         DW      DOLIT,100H,SLMOD
                DW      DOLIT,07H,ANDD
                DW      ROT,CAT,DOLIT,07H,SUBB
                DW      PLUS,ROT,CSTOR
                DW      SWAP,INCR,CSTOR
                DW      BRAN,FRRES7
FRRES4:         DW      DUPP,CAT,DOLIT,20H
                DW      ANDD,DOLIT,20H,EQUAL
                DW      QBRAN,FRRES5
                DW      DUPP,DOLIT,3,PLUS
                DW      HERE,SWAP,SUBB
                DW      SWAP,CELLP,CSTOR
                DW      BRAN,FRRES7
FRRES5:         DW      DUPP,CELLP,HERE
                DW      SWAP,SUBB,DUPP
                DW      ABSS,DOLIT,7FH,GREAT
                DW      QBRAN,FRRES6
                DW      ABORQ
                DB      13,' Out of range'
                DW      BRAN,FRRES7
FRRES6:         DW      SWAP,INCR,CSTOR
FRRES7:         DW      DOLIT,0,SWAP,STORE
                DW      BRAN,FRRES9
FRRES8:         DW      DROP
FRRES9:         DW      PLOP,FRRES1
                DW      EXIT

;   FRADD       ( fr# -- fr# )
;               Add forward reference to table

                $COLON  5,'FRADD',FRADD
                DW      DOLIT,0,DOLIT,0BH
                DW      TOR,TOR
FRADD1:         DW      DOLIT,0AH,RAT,EQUAL
                DW      ABORQ
                DB      27,'Too many forward references'
                DW      RAT,DOLIT,4,STAR
                DW      FRTABLE,PLUS,DUPP,AT
                DW      EQUAL_ZERO
                DW      QBRAN,FRADD2
                DW      DDUP,STORE,HERE
                DW      SWAP,CELLP,STORE
                DW      BRAN,FRADD4
FRADD2:         DW      DROP
                DW      RFROM,INCR,RAT,OVER,TOR,LESS
                DW      QBRAN,FRADD4
                DW      BRAN,FRADD1
FRADD4:         DW      RFROM,RFROM,DDROP,EXIT

;   !LTABLE     ( fr# -- )
;               Establish local label addr. value

                $COLON  7,'!LTABLE',SLTAB
                DW      FRRES,HERE,SWAP
                DW      DUPP,PLUS,LTABLE
                DW      PLUS,STORE
                DW      EXIT

;   @LTABLE     ( fr# -- addr.)
;               Fetch value of local table.If unfinished,Add current
;               location to forward ref. table

                $COLON  7,'@LTABLE',ATLTAB
                DW      DUPP,DUPP,PLUS
                DW      LTABLE,PLUS,AT
                DW      SWAP,OVER,EQUAL_ZERO
                DW      QBRAN,ATLTAB1
                DW      FRADD,DDROP,HERE
                DW      BRAN,ATLTAB2
ATLTAB1:        DW      DROP
ATLTAB2:        DW      EXIT

;; Local label definitions

                $COLON  3,'1$:',$1C
                DW      DOLIT,1,SLTAB,EXIT

                $COLON  3,'2$:',$2C
                DW      DOLIT,2,SLTAB,EXIT

                $COLON  3,'3$:',$3C
                DW      DOLIT,3,SLTAB,EXIT

                $COLON  3,'4$:',$4C
                DW      DOLIT,4,SLTAB,EXIT

                $COLON  3,'5$:',$5C
                DW      DOLIT,5,SLTAB,EXIT

                $COLON  3,'6$:',$6C
                DW      DOLIT,6,SLTAB,EXIT

                $COLON  3,'7$:',$7C
                DW      DOLIT,7,SLTAB,EXIT

                $COLON  3,'8$:',$8C
                DW      DOLIT,8,SLTAB,EXIT

                $COLON  3,'9$:',$9C
                DW      DOLIT,9,SLTAB,EXIT

                $COLON  3,'10$:',$10C
                DW      DOLIT,0AH,SLTAB,EXIT

;; Local label ferences

                $COLON  2,'1$',$1
                DW      DOLIT,1,ATLTAB,EXIT

                $COLON  2,'2$',$2
                DW      DOLIT,2,ATLTAB,EXIT

                $COLON  2,'3$',$3
                DW      DOLIT,3,ATLTAB,EXIT

                $COLON  2,'4$',$4
                DW      DOLIT,4,ATLTAB,EXIT

                $COLON  2,'5$',$5
                DW      DOLIT,5,ATLTAB,EXIT

                $COLON  2,'6$',$6
                DW      DOLIT,6,ATLTAB,EXIT

                $COLON  2,'7$',$7
                DW      DOLIT,7,ATLTAB,EXIT

                $COLON  2,'8$',$8
                DW      DOLIT,8,ATLTAB,EXIT

                $COLON  2,'9$',$9
                DW      DOLIT,9,ATLTAB,EXIT

                $COLON  2,'10$',$10
                DW      DOLIT,0AH,ATLTAB,EXIT

;; Error check

;   ?CONDITION  ( f -- )
;               Abort if f is false.

                $COLON  10,'?CONDITION',QCOND
                DW      INVER
                DW      ABORQ
                DB      19,' Conditionals Wrong'
                DW      EXIT

;; Equates to Addressing Modes

;   AMODE       ( -- )
;               Store addressing mode
;               0 = register direct     1 = immediate
;               2 = indirect            3 = indexed

                $USER   5,'AMODE',AMODE

;   AUTOINC     ( -- )
;               Store auto-incement flag
;               0 = don't auto-increment        1 = do auto-increment

                $USER   7,'AUTOINC',AUTOINC

;   SDMODE      ( -- )
;               Variable used for 3-operand instructions

                $USER   6,'SDMODE',SDMODE

;   MIOP        ( -- )
;               Variable

                $USER   4,'MIOP',MIOP

;; Assembler words for setting addressing mode

;   []          ( -- )
;               Indirect Addressing Mode

                $COLON  2,'[]',IDM
                DW      DOLIT,2,AMODE,STORE
                DW      EXIT

;   []+         ( -- )
;               Indirect autoincrement Addressing Mode

                $COLON  3,'[]+',IAM
                DW      DOLIT,2,AMODE,STORE
                DW      DOLIT,1,AUTOINC,STORE
                DW      EXIT

;   #           ( -- )
;               Immediate Addressing Mode

                $COLON  1,'#',IMM
                DW      DOLIT,1,AMODE,STORE
                DW      EXIT

;   [I]         ( -- )
;               Indexed Addressing Mode

                $COLON  3,'[I]',IXM
                DW      DOLIT,3,AMODE,STORE
                DW      EXIT

;   <--         ( -- )
;               3-operand instrution

                $COLON  3,'<--',ARROW
                DW      DOLIT,0FFFFH,SDMODE,STORE
                DW      EXIT

;   SD?C,       ( opcode -- )
;               Switch for 3-operand instructions

                $COLON  5,'SD?C,',SDQCCM
                DW      SDMODE,AT
                DW      QBRAN,SDQCCM1
                DW      MIOP,CAT,DOLIT,1CH,ORR
                DW      DOLIT,7CH,NEQUAL
                DW      QBRAN,SDQCCM2
                DW      ABORQ
                DB      29,' Not a 3-operand instruction '
                DW      BRAN,SDQCCM1
SDQCCM2:        DW      DOLIT,20H,XORR
SDQCCM1:        DW      CCOMMA
                DW      EXIT

;; Initialize all variables and flags

;   RESETF      ( -- )
;               Reset all flags

                $COLON  6,'RESETF',RESETF
                DW      DOLIT,0,AMODE,STORE
                DW      DOLIT,0,AUTOINC,STORE
                DW      DOLIT,0,SDMODE,STORE
                DW      EXIT

;; assembly words

;   BREG        ( breg -- )
;               Assemble a byte register reference

                $COLON  4,'BREG',BREG
                DW      DUPP,DOLIT,100H,LESS
                DW      QBRAN,BREG1
                DW      CCOMMA
                DW      BRAN,BREG2
BREG1:          DW      DROP,ABORQ
                DB      13,' reg over 255'
BREG2:          DW      EXIT

;   WREG        ( wreg -- )
;               Assemble a word register reference

                $COLON  4,'WREG',WREG
                DW      DUPP,DOLIT,1,ANDD,EQUAL_ZERO
                DW      QBRAN,WREG1
                DW      BREG
                DW      BRAN,WREG2
WREG1:          DW      DROP,ABORQ
                DB      13,'wreg not even'
WREG2:          DW      EXIT

;   LREG        ( lreg -- )
;               Assemble a long register reference

                $COLON  4,'LREG',LREG
                DW      DUPP,DOLIT,3,ANDD,EQUAL_ZERO
                DW      QBRAN,LREG1
                DW      BREG
                DW      BRAN,LREG2
LREG1:          DW      DROP,ABORQ
                DB      13,'  Lreg not /4'
LREG2:          DW      EXIT

;   OFFSET8,    ( A1 -- )
;               calculate 8-displace

                $COLON  8,'OFFSET8,',OFS8
                DW      HERE,INCR,SUBB
                DW      DUPP,ABSS,DOLIT,07FH,GREAT
                DW      QBRAN,OFS81
                DW      ABORQ
                DB      13,' Out of range'
                DW      BRAN,OFS82
OFS81:          DW      CCOMMA
OFS82:          DW      EXIT

;   OFFSET11,   ( opcode A1 -- )
;               calculate 11-displace

                $COLON  9,'OFFSET11,',OFS11
                DW      HERE,CELLP,SUBB
                DW      DUPP,ABSS,DOLIT,7FFH,GREAT
                DW      QBRAN,OFS111
                DW      ABORQ
                DB      13,' Out of range'
                DW      BRAN,OFS112
OFS111:         DW      DOLIT,100H,SLMOD,DOLIT,07H,ANDD
                DW      ROT,PLUS,CCOMMA,CCOMMA
OFS112:         DW      EXIT

;   OFFSET16,   ( A1 -- )
;               calculate 16-displace

                $COLON  9,'OFFSET16,',OFS16
                DW      HERE,CELLP,SUBB,COMMA
                DW      EXIT


;; Words to build the instructions:

;   1MI         ( opcode -- )
;               Single byte inst.

                $COLON  3,'1MI',MI1
                DW      CCOMMA
                DW      RESETF
                DW      EXIT

; --------------- 1MI inst. -----------

;   CLRC        ( -- )

                $COLON  4,'CLRC',CLRC
                DW      DOLIT,0F8H,MI1
                DW      EXIT

;   CLRVT       ( -- )

                $COLON  5,'CLRVT',CLRVT
                DW      DOLIT,0FCH,MI1
                DW      EXIT

;   DI          ( -- )

                $COLON  2,'DI',DDII
                DW      DOLIT,0FAH,MI1
                DW      EXIT

;   EI          ( -- )

                $COLON  2,'EI',EEII
                DW      DOLIT,0FBH,MI1
                DW      EXIT

;   NOP         ( -- )

                $COLON  3,'NOP',NOP
                DW      DOLIT,0FDH,MI1
                DW      EXIT

;   POPF        ( -- )

                $COLON  4,'POPF',POPF
                DW      DOLIT,0F3H,MI1
                DW      EXIT

;   PUSHF       ( -- )

                $COLON  5,'PUSHF',PUSHF
                DW      DOLIT,0F2H,MI1
                DW      EXIT

;   RET         ( -- )

                $COLON  3,'RET',RET
                DW      DOLIT,0F0H,MI1
                DW      EXIT

;   RST         ( -- )

                $COLON  3,'RST',RST
                DW      DOLIT,0FFH,MI1
                DW      EXIT

;   SETC        ( -- )

                $COLON  4,'SETC',SETC
                DW      DOLIT,0F9H,MI1
                DW      EXIT

;   TRAP        ( -- )

                $COLON  4,'TRAP',TRAP
                DW      DOLIT,0F7H,MI1
                DW      EXIT

; -----------------------------------------

;   2MI         ( breg opcode -- )
;               e.g. breg CLRB

                $COLON  3,'2MI',MI2
                DW      CCOMMA,BREG
                DW      RESETF
                DW      EXIT

; ------------------ 2MI inst. -------------

;   CLRB        ( breg -- )

                $COLON  4,'CLRB',CLRB
                DW      DOLIT,11H,MI2
                DW      EXIT

;   DECB        ( breg -- )

                $COLON  4,'DECB',DECB
                DW      DOLIT,15H,MI2
                DW      EXIT

;   INCB        ( breg -- )

                $COLON  4,'INCB',INCB
                DW      DOLIT,17H,MI2
                DW      EXIT

;   NEGB        ( breg -- )

                $COLON  4,'NEGB',NEGB
                DW      DOLIT,13H,MI2
                DW      EXIT

;   NOTB        ( breg -- )

                $COLON  4,'NOTB',NOTB
                DW      DOLIT,12H,MI2
                DW      EXIT

;   SKIP        ( breg -- )

                $COLON  4,'SKIP',SKIP
                DW      DOLIT,0,MI2
                DW      EXIT

; ----------------------------------------

;   3MI         ( wreg opcode -- )
;               e.g.  wreg CLR

                $COLON  3,'3MI',MI3
                DW      CCOMMA,WREG
                DW      RESETF
                DW      EXIT

; ----------------- 3MI inst. --------------

;   BR[]        ( wreg -- )

                $COLON  4,'BR[]',BRR
                DW      DOLIT,0E3H,MI3
                DW      EXIT

;   CLR         ( wreg -- )

                $COLON  3,'CLR',CLR
                DW      DOLIT,01H,MI3
                DW      EXIT

;   DEC         ( wreg -- )

                $COLON  3,'DEC',DEC
                DW      DOLIT,05H,MI3
                DW      EXIT

;   EXTB        ( wreg -- )

                $COLON  4,'EXTB',EXTB
                DW      DOLIT,16H,MI3
                DW      EXIT

;   INC         ( wreg -- )

                $COLON  3,'INC',INC
                DW      DOLIT,07H,MI3
                DW      EXIT

;   NEG         ( wreg -- )

                $COLON  3,'NEG',NEG
                DW      DOLIT,03H,MI3
                DW      EXIT

;   NOT         ( wreg -- )

                $COLON  3,'NOT',NOT
                DW      DOLIT,02H,MI3
                DW      EXIT

; ---------------------------------------

;   4MI         ( Lreg cnt/breg opcode -- )
;               e.g. Lreg cnt/breg SHLL

                $COLON  3,'4MI',MI4
                DW      CCOMMA,BREG,LREG
                DW      RESETF
                DW      EXIT

; ---------------- 4MI inst,-------------

;   NORML       ( Lreg breg -- )

                $COLON  5,'NORML',NORML
                DW      DOLIT,0FH,MI4
                DW      EXIT

;   SHLL        ( Lreg breg -- )

                $COLON  4,'SHLL',SHLL
                DW      DOLIT,0DH,MI4
                DW      EXIT

;   SHRAL       ( Lreg breg -- )

                $COLON  5,'SHRAL',SHRAL
                DW      DOLIT,0EH,MI4
                DW      EXIT

;   SHRL        ( Lreg breg -- )

                $COLON  4,'SHRL',SHRL
                DW      DOLIT,0CH,MI4
                DW      EXIT

; ----------------------------------------

;   5MI         ( breg cnt/breg opcode -- )
;               e.g. breg cnt/breg SHLL

                $COLON  3,'5MI',MI5
                DW      CCOMMA,BREG,BREG
                DW      RESETF
                DW      EXIT

; ----------------- 5MI inst.--------------

;   SHLB        ( breg breg -- )

                $COLON  4,'SHLB',SHLB
                DW      DOLIT,19H,MI5
                DW      EXIT

;   SHRAB       ( breg breg -- )

                $COLON  5,'SHRAB',SHRAB
                DW      DOLIT,1AH,MI5
                DW      EXIT

;   SHRB        ( breg breg -- )

                $COLON  4,'SHRB',SHRB
                DW      DOLIT,18H,MI5
                DW      EXIT

; ----------------------------------------

;   6MI         ( wreg cnt/breg opcode -- )
;               e.g. wreg cnt/breg SHLL

                $COLON  3,'6MI',MI6
                DW      CCOMMA,BREG,WREG
                DW      RESETF
                DW      EXIT

; ----------------- 6MI inst. -------------

;   SHL         ( wreg breg -- )

                $COLON  3,'SHL',SHL
                DW      DOLIT,09H,MI6
                DW      EXIT

;   SHR         ( wreg breg -- )

                $COLON  3,'SHR',SHR
                DW      DOLIT,08H,MI6
                DW      EXIT

;   SHRA        ( wreg Breg -- )

                $COLON  4,'SHRA',SHRA
                DW      DOLIT,0AH,MI6
                DW      EXIT

; ------------------------------------------

;   7MI         ( waop  opcode -- ) \ e.g. PUSH   POP
;               ( offset wreg  opcode -- ) \ indexed

                $COLON  3,'7MI',MI7
                DW      MIOP,STORE
                DW      AMODE,AT
                DW      DUPP,EQUAL_ZERO
                DW      QBRAN,MI7_1
                DW      DROP,MIOP,AT,CCOMMA,WREG        ; register direct
                DW      BRAN,MI7_5
MI7_1:          DW      DUPP,DOLIT,1,EQUAL
                DW      QBRAN,MI7_2
                DW      DROP,MIOP,AT
                DW      DOLIT,1,ORR,CCOMMA,COMMA        ; immediate
                DW      BRAN,MI7_5
MI7_2:          DW      DUPP,DOLIT,2,EQUAL
                DW      QBRAN,MI7_3
                DW      DROP,MIOP,AT                    ; indirect
                DW      DOLIT,2,ORR,CCOMMA
                DW      DOLIT,0FEH,ANDD
                DW      AUTOINC,AT,ORR,CCOMMA
                DW      BRAN,MI7_5
MI7_3:          DW      DROP,MIOP,AT                    ; indexed
                DW      DOLIT,3,ORR,CCOMMA
                DW      OVER,DOLIT,100H,ULESS
                DW      QBRAN,MI7_4
                DW      DOLIT,0FEH,ANDD,CCOMMA,CCOMMA   ; -short
                DW      BRAN,MI7_5
MI7_4:          DW      DOLIT,1,ORR,CCOMMA,COMMA        ; -long
MI7_5:          DW      RESETF,EXIT

; ------------------- 7MI inst.--------------------

;   PUSH        ( waop -- ) or ( offset wreg -- )

                $COLON  4,'PUSH',PUSH
                DW      DOLIT,0C8H,MI7
                DW      EXIT

;   POP         ( waop -- ) or ( offset wreg -- )

                $COLON  3,'POP',POP
                DW      DOLIT,0CCH,MI7
                DW      EXIT

; -----------------------------------------------

;   8MI         ( breg baop opcode -- ) \ e.g.  breg baop ADDB
;               ( breg offset wreg opcode -- ) \ indexed
;               ( Dbreg Sbreg baop opcode -- ) \ e.g.Dbreg <-- Sbreg baop ADDB

                $COLON  3,'8MI',MI8
                DW      MIOP,STORE
                DW      AMODE,AT
                DW      DUPP,EQUAL_ZERO
                DW      QBRAN,MI8_1
                DW      DROP,MIOP,AT                    ; register direct
                DW      SDQCCM,BREG
                DW      BRAN,MI8_5
MI8_1:          DW      DUPP,DOLIT,1,EQUAL
                DW      QBRAN,MI8_2
                DW      DROP,MIOP,AT                    ; immediate
                DW      DOLIT,1,ORR,SDQCCM,CCOMMA
                DW      BRAN,MI8_5
MI8_2:          DW      DUPP,DOLIT,2,EQUAL
                DW      QBRAN,MI8_3
                DW      DROP,MIOP,AT                    ; indirect
                DW      DOLIT,2,ORR,SDQCCM
                DW      DOLIT,0FEH,ANDD,AUTOINC,AT
                DW      ORR,CCOMMA
                DW      BRAN,MI8_5
MI8_3:          DW      DROP,MIOP,AT                    ; indexed
                DW      DOLIT,3,ORR,SDQCCM
                DW      OVER,DOLIT,100H,ULESS
                DW      QBRAN,MI8_4
                DW      DOLIT,0FEH,ANDD,CCOMMA,CCOMMA   ; -short
                DW      BRAN,MI8_5
MI8_4:          DW      DOLIT,1,ORR,CCOMMA,COMMA        ; -long
MI8_5:          DW      BREG,SDMODE,AT
                DW      QBRAN,MI8_6
                DW      BREG
MI8_6:          DW      RESETF
                DW      EXIT

; ---------------------- 8MI inst.---------------------------

;   ADDB        ( -- )

                $COLON  4,'ADDB',ADDB
                DW      DOLIT,74H,MI8
                DW      EXIT

;   ADDCB       ( -- )

                $COLON  5,'ADDCB',ADDCB
                DW      DOLIT,0B4H,MI8
                DW      EXIT

;   ANDB        ( -- )

                $COLON  4,'ANDB',ANDB
                DW      DOLIT,70H,MI8
                DW      EXIT

;   CMPB        ( -- )

                $COLON  4,'COMPB',COMPB
                DW      DOLIT,98H,MI8
                DW      EXIT

;   LDB         ( -- )

                $COLON  3,'LDB',LDB
                DW      DOLIT,0B0H,MI8
                DW      EXIT

;   ORB         ( -- )

                $COLON  3,'ORB',ORB
                DW      DOLIT,90H,MI8
                DW      EXIT

;   STB         ( -- )

                $COLON  3,'STB',STB
                DW      DOLIT,0C4H,MI8
                DW      EXIT

;   SUBB        ( -- )

                $COLON  4,'SUBB',SUBBB
                DW      DOLIT,78H,MI8
                DW      EXIT

;   SUBCB       ( -- )

                $COLON  5,'SUBCB',SUBCB
                DW      DOLIT,0B8H,MI8
                DW      EXIT

;   XORB        ( -- )

                $COLON  4,'XORB',XORB
                DW      DOLIT,94H,MI8
                DW      EXIT

; ----------------------------------------------

;   9MI         ( wreg waop opcode -- ) \ e.g.  wreg waop ADD
;               ( wreg offset wreg opcode -- ) \ indexed
;               ( Dwreg Swreg waop opcode -- ) \ e.g.Dwreg <-- Sweg wop ADD

                $COLON  3,'9MI',MI9
                DW      MIOP,STORE
                DW      AMODE,AT
                DW      DUPP,EQUAL_ZERO
                DW      QBRAN,MI9_1
                DW      DROP,MIOP,AT                    ; register direct
                DW      SDQCCM,WREG
                DW      BRAN,MI9_5
MI9_1:          DW      DUPP,DOLIT,1,EQUAL
                DW      QBRAN,MI9_2
                DW      DROP,MIOP,AT                    ; immediate
                DW      DOLIT,1,ORR,SDQCCM,COMMA
                DW      BRAN,MI9_5
MI9_2:          DW      DUPP,DOLIT,2,EQUAL
                DW      QBRAN,MI9_3
                DW      DROP,MIOP,AT                    ; indirect
                DW      DOLIT,2,ORR,SDQCCM
                DW      DOLIT,0FEH,ANDD,AUTOINC,AT
                DW      ORR,CCOMMA
                DW      BRAN,MI9_5
MI9_3:          DW      DROP,MIOP,AT                    ; indexed
                DW      DOLIT,3,ORR,SDQCCM
                DW      OVER,DOLIT,100H,ULESS
                DW      QBRAN,MI9_4
                DW      DOLIT,0FEH,ANDD,CCOMMA,CCOMMA   ; -short
                DW      BRAN,MI9_5
MI9_4:          DW      DOLIT,1,ORR,CCOMMA,COMMA        ; -long
MI9_5:          DW      WREG,SDMODE,AT
                DW      QBRAN,MI9_6
                DW      WREG
MI9_6:          DW      RESETF
                DW      EXIT

; -------------------- 9MI inst.--------------------------

;   ADD         ( -- )

                $COLON  3,'ADD',ADD
                DW      DOLIT,64H,MI9
                DW      EXIT

;   ADDC        ( -- )

                $COLON  4,'ADDC',ADDC
                DW      DOLIT,0A4H,MI9
                DW      EXIT

;   AND         ( -- )

                $COLON  3,'AND',AND
                DW      DOLIT,60H,MI9
                DW      EXIT

;   CMP         ( -- )

                $COLON  3,'CMP',CMP
                DW      DOLIT,88H,MI9
                DW      EXIT

;   LD          ( -- )

                $COLON  2,'LD',LD
                DW      DOLIT,0A0H,MI9
                DW      EXIT

;   OR          ( -- )

                $COLON  2,'OR',OR
                DW      DOLIT,80H,MI9
                DW      EXIT

;   ST          ( -- )

                $COLON  2,'ST',SSTT
                DW      DOLIT,0C0H,MI9
                DW      EXIT

;   SUB         ( -- )

                $COLON  3,'SUB',SUB
                DW      DOLIT,68H,MI9
                DW      EXIT

;   SUBC        ( -- )

                $COLON  4,'SUBC',SUBC
                DW      DOLIT,0A8H,MI9
                DW      EXIT

;  XOR          ( -- )

                $COLON  3,'XOR',XOR
                DW      DOLIT,84H,MI9
                DW      EXIT

; -------------------------------------

;   10MI        ( Lreg waop opcode -- ) \ e.g.  Lreg waop DIVU
;               ( Lreg offset wreg opcode -- ) \ indexed
;               ( Lreg wreg waop opcode -- ) \ e.g.Lreg <-- weg wop MUL

                $COLON  4,'10MI',MI10
                DW      MIOP,STORE
                DW      AMODE,AT
                DW      DUPP,EQUAL_ZERO
                DW      QBRAN,MI10_1
                DW      DROP,MIOP,AT                    ; register direct
                DW      SDQCCM,WREG
                DW      BRAN,MI10_5
MI10_1:         DW      DUPP,DOLIT,1,EQUAL
                DW      QBRAN,MI10_2
                DW      DROP,MIOP,AT                    ; immediate
                DW      DOLIT,1,ORR,SDQCCM,COMMA
                DW      BRAN,MI10_5
MI10_2:         DW      DUPP,DOLIT,2,EQUAL
                DW      QBRAN,MI10_3
                DW      DROP,MIOP,AT                    ; indirect
                DW      DOLIT,2,ORR,SDQCCM
                DW      DOLIT,0FEH,ANDD,AUTOINC,AT
                DW      ORR,CCOMMA
                DW      BRAN,MI10_5
MI10_3:         DW      DROP,MIOP,AT                    ; indexed
                DW      DOLIT,3,ORR,SDQCCM
                DW      OVER,DOLIT,100H,ULESS
                DW      QBRAN,MI10_4
                DW      DOLIT,0FEH,ANDD,CCOMMA,CCOMMA   ; -short
                DW      BRAN,MI10_5
MI10_4:         DW      DOLIT,1,ORR,CCOMMA,COMMA        ; -long
MI10_5:         DW      SDMODE,AT
                DW      QBRAN,MI10_6
                DW      WREG
MI10_6:         DW      LREG,RESETF
                DW      EXIT

; ------------------- 10MI inst.---------------------

;   DIVU        ( -- )

                $COLON  4,'DIVU',DIVU
                DW      DOLIT,8CH,MI10
                DW      EXIT

;   MULU        ( -- )

                $COLON  4,'MULU',MULU
                DW      DOLIT,6CH,MI10
                DW      EXIT

;   DIV         ( -- )

                $COLON  3,'DIV',DIV
                DW      DOLIT,0FEH,CCOMMA
                DW      DIVU,EXIT

;   MUL         ( -- )

                $COLON  3,'MUL',MUL
                DW      DOLIT,0FEH,CCOMMA
                DW      MULU,EXIT

; -------------------------------------------

;   11MI        ( wreg baop opcode -- ) \ e.g.  wreg baop DIVUB
;               ( wreg offset wreg opcode -- ) \ indexed
;               ( wreg breg baop opcode -- ) \ e.g.wreg <-- beg baop MULB

                $COLON  4,'11MI',MI11
                DW      MIOP,STORE
                DW      AMODE,AT
                DW      DUPP,EQUAL_ZERO
                DW      QBRAN,MI11_1
                DW      DROP,MIOP,AT                    ; register direct
                DW      SDQCCM,BREG
                DW      BRAN,MI11_5
MI11_1:         DW      DUPP,DOLIT,1,EQUAL
                DW      QBRAN,MI11_2
                DW      DROP,MIOP,AT                    ; immediate
                DW      DOLIT,1,ORR,SDQCCM,COMMA
                DW      BRAN,MI11_5
MI11_2:         DW      DUPP,DOLIT,2,EQUAL
                DW      QBRAN,MI11_3
                DW      DROP,MIOP,AT                    ; indirect
                DW      DOLIT,2,ORR,SDQCCM
                DW      DOLIT,0FEH,ANDD,AUTOINC,AT
                DW      ORR,CCOMMA
                DW      BRAN,MI11_5
MI11_3:         DW      DROP,MIOP,AT                    ; indexed
                DW      DOLIT,3,ORR,SDQCCM
                DW      OVER,DOLIT,100H,ULESS
                DW      QBRAN,MI11_4
                DW      DOLIT,0FEH,ANDD,CCOMMA,CCOMMA   ; -short
                DW      BRAN,MI11_5
MI11_4:         DW      DOLIT,1,ORR,CCOMMA,COMMA        ; -long
MI11_5:         DW      SDMODE,AT
                DW      QBRAN,MI11_6
                DW      BREG
MI11_6:         DW      WREG,RESETF
                DW      EXIT

; ---------------------- 11MI inst.---------------

;   DIVUB       ( -- )

                $COLON  5,'DIVUB',DIVUB
                DW      DOLIT,9CH,MI11
                DW      EXIT

;   MULUB       ( -- )

                $COLON  5,'MULUB',MULUB
                DW      DOLIT,7CH,MI11
                DW      EXIT

;   LDBZE       ( -- )

                $COLON  5,'LDBZE',LDBZE
                DW      DOLIT,0ACH,MI11
                DW      EXIT

;   LDBSE       ( -- )

                $COLON  5,'LDBSE',LDBSE
                DW      DOLIT,0BCH,MI11
                DW      EXIT

;   DIVB        ( -- )

                $COLON  4,'DIVB',DIVB
                DW      DOLIT,0FEH,CCOMMA
                DW      DIVUB,EXIT

;   MULB        ( -- )

                $COLON  4,'MULB',MULB
                DW      DOLIT,0FEH,CCOMMA
                DW      MULUB,EXIT

; ---------------------------------------------

;   12MI        ( A1 opcode  -- )
;               e.g. LJMP    LCALL

                $COLON  4,'12MI',MI12
                DW      CCOMMA,OFS16,RESETF
                DW      EXIT

; -------------------- 12MI inst.--------------

;   LJMP        ( A1 -- )

                $COLON  4,'LJMP',LJMP
                DW      DOLIT,0E7H,MI12
                DW      EXIT

;   LCALL       ( A1 -- )

                $COLON  5,'LCALL',LCALL
                DW      DOLIT,0EFH,MI12
                DW      EXIT

; ----------------------------------------------

;   13MI        ( A1 opcode -- )
;               e.g. SJMP  SCALL

                $COLON  4,'13MI',MI13
                DW      SWAP,OFS11,RESETF
                DW      EXIT

; ----------------------- 13MI inst.------------

;   SJMP        ( A1 -- )

                $COLON  4,'SJMP',SJMP
                DW      DOLIT,20H,MI13
                DW      EXIT

;   SCALL       ( A1 -- )

                $COLON  5,'SCALL',SCALL
                DW      DOLIT,28H,MI13
                DW      EXIT

; ----------------------------------------------

;   14MI        ( A1 opcode -- )

                $COLON  4,'14MI',MI14
                DW      CCOMMA,OFS8,RESETF
                DW      EXIT

; ------------------- 14MI inst.----------------

;   JC          ( A1 -- )

                $COLON  2,'JC',JC
                DW      DOLIT,0DBH,MI14
                DW      EXIT

;   JNC         ( A1 -- )

                $COLON  3,'JNC',JNC
                DW      DOLIT,0D3H,MI14
                DW      EXIT

;   JH          ( A1-- )

                $COLON  2,'JH',JH
                DW      DOLIT,0D9H,MI14
                DW      EXIT

;   JNH         ( A1 -- )

                $COLON  3,'JNH',JNH
                DW      DOLIT,0D1H,MI14
                DW      EXIT

;   JGT         ( A1 -- )

                $COLON  3,'JGT',JGT
                DW      DOLIT,0D2H,MI14
                DW      EXIT

;   JLE         ( A1 -- )

                $COLON  3,'JLE',JLE
                DW      DOLIT,0DAH,MI14
                DW      EXIT

;   JE          ( A1 -- )

                $COLON  2,'JE',JE
                DW      DOLIT,0DFH,MI14
                DW      EXIT

;   JNE         ( A1 -- )

                $COLON  3,'JNE',JNE
                DW      DOLIT,0D7H,MI14
                DW      EXIT

;   JV          ( A1 -- )

                $COLON  2,'JV',JV
                DW      DOLIT,0DDH,MI14
                DW      EXIT

;   JNV         ( A1 -- )

                $COLON  3,'JNV',JNV
                DW      DOLIT,0D5H,MI14
                DW      EXIT

;   JST         ( A1 -- )

                $COLON  3,'JST',JST
                DW      DOLIT,0D8H,MI14
                DW      EXIT

;   JNST        ( A1 -- )

                $COLON  4,'JNST',JNST
                DW      DOLIT,0D0H,MI14
                DW      EXIT

;   JGE         ( A1 -- )

                $COLON  3,'JGE',JGE
                DW      DOLIT,0D6H,MI14
                DW      EXIT

;   JLT         ( A1 -- )

                $COLON  3,'JLT',JLT
                DW      DOLIT,0DEH,MI14
                DW      EXIT

;   JVT         ( A1 -- )

                $COLON  3,'JVT',JVT
                DW      DOLIT,0DCH,MI14
                DW      EXIT

;   JNVT        ( A1 -- )

                $COLON  4,'JNVT',JNVT
                DW      DOLIT,0D4H,MI14
                DW      EXIT

; ----------------------------------------

;   15MI       ( breg cnt A1 opcode -- )
;               e.g. JBS JBC

                $COLON  4,'15MI',MI15
                DW      ROT,PLUS,CCOMMA,SWAP,CCOMMA
                DW      OFS8,RESETF,EXIT

; -------------------15MI inst.-------------

;   JBS         ( breg cnt A1 -- )

                $COLON  3,'JBS',JBS
                DW      DOLIT,38H,MI15
                DW      EXIT

;   JBC         ( breg cnt A1 -- )

                $COLON  3,'JBC',JBC
                DW      DOLIT,30H,MI15
                DW      EXIT

; ------------------------------------------

;   DJNZ        ( breg A1 -- )

                $COLON  4,'DJNZ',DJNZ
                DW      DOLIT,0E0H,CCOMMA
                DW      SWAP,CCOMMA,OFS8
                DW      RESETF,EXIT


;; Follow is memery relocation.

CODDEE = 1C00H
ORG     CODDEE

;; Jump instructions for IF ELSE THEN tec.

;   BITSET      ( breg n -- )
;               e.g. breg n BITSET IF

                $COLON  6,'BITSET',BITSET
                DW      DOLIT,7,ANDD
                DW      DOLIT,30H,ORR
                DW      CCOMMA,BREG,RESETF
                DW      EXIT

;   BITCLEAR    ( breg n -- )
;               e.g.  breg n BITCLEAR IF

                $COLON  8,'BITCLEAR',BITCLEAR
                DW      DOLIT,7,ANDD
                DW      DOLIT,38H,ORR
                DW      CCOMMA,BREG,RESETF
                DW      EXIT

;   DEC0=       ( breg -- )
;               e.g. breg DEC0= UNTIL

                $COLON  5,'DEC0=',DECZERO
                DW      DOLIT,0E0H,CCOMMA,BREG,RESETF
                DW      EXIT

;; The conditional jumps for assembler IF , BEGIN , etc.

;   ST0<>       (  -- )
;               \  JNST

                $COLON  5,'ST0<>',STNZ
                DW      DOLIT,0D0H,MI1
                DW      EXIT

;   ST0=        ( -- )
;               \ JST

                $COLON  4,'ST0=',STZ
                DW      DOLIT,0D8H,MI1
                DW      EXIT

;   U0>         ( -- )
;               \ JNH

                $COLON  3,'U0>',UZGT
                DW      DOLIT,0D1H,MI1
                DW      EXIT

;   U0<=        ( -- )
;               \ JH

                $COLON  5,'U0<=',UZLE
                DW      DOLIT,0D9H,MI1
                DW      EXIT

;   0<=         ( -- )
;               \  JGT

                $COLON  3,'0<=',ZLTE
                DW      DOLIT,0D2H,MI1
                DW      EXIT

;   0>          ( -- )
;               \ JLE

                $COLON  2,'0>',ZGT
                DW      DOLIT,0DAH,MI1
                DW      EXIT

;   C0<>        ( -- )
;               \ JNC

                $COLON  4,'C0<>',CNZ
                DW      DOLIT,0D3H,MI1
                DW      EXIT

;   C0=         ( -- )
;               \ JC

                $COLON  3,'C0=',CZ
                DW      DOLIT,0DBH,MI1
                DW      EXIT

;   VT0<>       ( -- )
;               \  JNVT

                $COLON  5,'VT0<>',VTNZ
                DW      DOLIT,0D4H,MI1
                DW      EXIT

;   VT0=        (  opcode -- )
;               \ JVT

                $COLON  4,'VT0=',VTZ
                DW      DOLIT,0DCH,MI1
                DW      EXIT

;   V0<>        ( -- )
;               \  JNV

                $COLON  4,'V0<>',VNZ
                DW      DOLIT,0D5H,MI1
                DW      EXIT

;   V0=         ( -- )
;               \ JV

                $COLON  3,'V0=',VZ
                DW      DOLIT,0DDH,MI1
                DW      EXIT

;   0<          ( -- )
;               \ JGE

                $COLON  2,'0<',ZLT
                DW      DOLIT,0D6H,MI1
                DW      EXIT

;   0>=         ( -- )
;               \ JLT

                $COLON  3,'0>=',ZGTE
                DW      DOLIT,0DEH,MI1
                DW      EXIT

;   0=          ( -- )
;               \ JNE

                $COLON  2,'0=',ZER
                DW      DOLIT,0D7H,MI1
                DW      EXIT

;   0<>         ( -- )
;               \ JE

                $COLON  3,'0<>',NZER
                DW      DOLIT,0DFH,MI1
                DW      EXIT

;   UJMP>       ( -- )
;               SJMP forwards up to 127

                $COLON  5,'UJMP>',UJMPT
                DW      DOLIT,20H,MI1
                DW      EXIT

;   <UJMP       ( -- )
;               SJMP backwards up to 128

                $COLON  5,'<UJMP',LUJMP
                DW      DOLIT,27H,MI1
                DW      EXIT

;   X?>MARK     ( -- f addr )

                $COLON  7,'X?>MARK',XQTMK
                DW      DOLIT,0FFFFH,HERE
                DW      DOLIT,0,CCOMMA
                DW      EXIT

;   X?>RESOLVE  ( f addr -- )

                $COLON  10,'X?>RESOLVE',XQTRS
                DW      HERE,OVER,INCR,SUBB
                DW      SWAP,CSTOR,QCOND
                DW      EXIT

;   X?<MARK     ( -- f addr )

                $COLON  7,'X?<MARK',XQLMK
                DW      DOLIT,0FFFFH,HERE
                DW      EXIT

;   X?<RESOLVE  ( f addr -- )

                $COLON  10,'X?<RESOLVE',XQLRS
                DW      HERE,INCR,SUBB
                DW      CCOMMA,QCOND
                DW      EXIT

;   BEGIN       ( -- a f )

                $COLON  5,'BEGIN',ABEGIN
                DW      XQLMK,EXIT

;   UNTIL       ( a f -- )

                $COLON  5,'UNTIL',AUNTIL
                DW      XQLRS,EXIT

;   AGAIN       ( a f -- )
;               SJMP backwards up to 127

                $COLON  5,'AGAIN',AAGAIN
                DW      LUJMP,AUNTIL,EXIT

;   IF          ( -- A f )

                $COLON  2,'IF',AIF
                DW      XQTMK,EXIT

;   THEN        ( A f -- )

                $COLON  4,'THEN',ATHEN
                DW      XQTRS,EXIT

;   ELSE        ( A f -- A f )
;               SJMP forwards up to 127

                $COLON  4,'ELSE',AELSE
                DW      UJMPT,AIF,DSWAP,ATHEN,EXIT

;   REPEAT      ( A f a f -- )

                $COLON  6,'REPEAT',AREPEAT
                DW      AAGAIN,ATHEN,EXIT

;   WHILE       ( a f -- A f a f )

                $COLON  5,'WHILE',AWHILE
                DW      AIF,DSWAP,EXIT

;; symbolic names for the i/o registers of the 8096/8098.

;   R0          ( -- )

                $CONST  2,'R0',R0
                DW      0

;   AD_COMMAND  ( -- )

                $CONST  10,'AD_COMMAND',ADCOM
                DW      2

;   AD_RESULT   ( -- )

                $CONST  9,'AD_RESULT',ADRT
                DW      2

;   HSI_MODE    ( -- )

                $CONST  8,'HSI_MODE',HSIMD
                DW      3

;   HSI_TIME    ( -- )

                $CONST  8,'HSI_TIME',HSITM
                DW      4

;   HSO_TIME    ( -- )

                $CONST  8,'HSO_TIME',HSOTM
                DW      4

;   HSO_COMMAND ( -- )

                $CONST  11,'HSO_COMMAND',HSOCOM
                DW      6

;   HSI_STATUS  ( -- )

                $CONST  10,'HSI_STATUS',HSIST
                DW      6

;   SBUF        ( -- )

                $CONST  4,'SBUF',SBUF
                DW      7

;   INT_MASK    ( -- )

                $CONST  8,'INT_MASK',INTMK
                DW      8

;   INT_PEND    ( -- )

                $CONST  8,'INT_PEND',INTPD
                DW      9

;   TIMER1      ( -- )

                $CONST  6,'TIMER1',TIMER1
                DW      0AH

;   WATCHDOG    ( -- )

                $CONST  8,'WATCHDOG',WATCHDG
                DW      0AH

;   TIMER2      ( -- )

                $CONST  6,'TIMER2',TIMER2
                DW      0CH

;   BAUD_RATE   ( -- )

                $CONST  9,'BAUD_RATE',BAUDRT
                DW      0EH

;   IOPORT0     ( -- )

                $CONST  7,'IOPORT0',PORT0
                DW      0EH

;   IOPORT2     ( -- )

                $CONST  7,'IOPORT2',PORT2
                DW      10H

;   SP_CON      ( -- )

                $CONST  6,'SP_CON',SPCON
                DW      11H

;   SP_STAT     ( -- )

                $CONST  7,'SP_STAT',SPSTAT
                DW      11H

;   IOC0        ( -- )

                $CONST  4,'IOC0',IOC0
                DW      15H

;   IOS0        ( -- )

                $CONST  4,'IOS0',IOS0
                DW      15H

;   IOC1        ( -- )

                $CONST  4,'IOC1',IOC1
                DW      16H

;   IOS1        ( -- )

                $CONST  4,'IOS1',IOS1
                DW      16H

;   PWM_CONTROL ( -- )

                $CONST  11,'PWM_CONTROL',PWMCNL
                DW      17H

;   NEXT        ( -- )
;               Assembler in-line $NEXT.

                $COLON  4,'NEXT',ANEXT
                DW      DOLIT,20H,DOLIT,1AH
                DW      IAM,LD                  ; AX, IP []+ LD
                DW      DOLIT,20H,BRR           ; AX BR[]
                DW      EXIT

;   END-CODE    ( -- )
;               Terminate CODE - def.

                $COLON  IMEDD+8,'END-CODE',ENDCODE
                DW      FRCLR,RESETF,QCSP
                DW      FORTH,EXIT

;; ++++++++++++++++++++++++++++++++++++++++++++++++++++

ALASTN          EQU     _NAME+4                 ;last name address in name dictionary

NTOP            EQU     NNPP                    ;next available memory in name dictionary
CTOP		EQU	CCPP			;next available memory in code dictionary

MAIN		ENDS

END	ORIG
